import asyncio
import datetime
import os
import webbrowser
import math
import platform
import sys
import subprocess
from urllib.request import url2pathname, pathname2url
from io import StringIO

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import Gtk
from gi.repository import GLib

from pychess.System.Log import log
from pychess.System import conf, uistuff, prefix
from pychess.Utils.const import HINT, NAME, SPY, NORMALCHESS
from pychess.Utils.checkversion import isgit, checkversion
from pychess.Utils.IconLoader import get_pixbuf
from pychess.widgets import enginesDialog, newGameDialog
from pychess.widgets.Background import hexcol
from pychess.widgets.tipOfTheDay import TipOfTheDay
from pychess.widgets.discovererDialog import DiscovererDialog
from pychess.widgets.ExternalsDialog import ExternalsDialog
from pychess.widgets import gamewidget
from pychess.widgets.analyzegameDialog import AnalyzeGameDialog
from pychess.widgets import preferencesDialog, gameinfoDialog, playerinfoDialog
from pychess.widgets.TaskerManager import internet_game_tasker
from pychess.widgets.RecentChooser import recent_menu, recent_manager
from pychess.widgets.prompttext import getUserTextDialog
from pychess.Players.engineNest import discoverer
from pychess.Savers import chesspastebin
from pychess.Savers.pgn import PGNFile
from pychess.Savers.remotegame import get_internet_game
from pychess.System.protoopen import splitUri
from pychess.widgets import mainwindow
from pychess.ic import ICLogon
from pychess.perspectives import perspective_manager
from pychess.perspectives.welcome import Welcome
from pychess.perspectives.games import Games, get_open_dialog
from pychess.perspectives.learn import Learn
from pychess.perspectives.fics import FICS
from pychess.perspectives.database import Database
from pychess import VERSION, VERSION_NAME

leftkeys = list(map(Gdk.keyval_from_name, ("Left", "KP_Left")))
rightkeys = list(map(Gdk.keyval_from_name, ("Right", "KP_Right")))
upkeys = list(map(Gdk.keyval_from_name, ("Up", "KP_Up")))
downkeys = list(map(Gdk.keyval_from_name, ("Down", "KP_Down")))
homekeys = list(map(Gdk.keyval_from_name, ("Home", "KP_Home")))
endkeys = list(map(Gdk.keyval_from_name, ("End", "KP_End")))
functionkeys = [
    Gdk.keyval_from_name(k)
    for k in ("F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8", "F9", "F10", "F11")
]

TARGET_TYPE_URI_LIST = 0xBADBEEF
DRAG_ACTION = Gdk.DragAction.COPY
DRAG_RESTRICT = Gtk.TargetFlags.OTHER_APP
DND_LIST = [Gtk.TargetEntry.new("text/uri-list", DRAG_RESTRICT, TARGET_TYPE_URI_LIST)]


class GladeHandlers:
    def __init__(self, app):
        self.app = app

    def on_window_key_press(self, window, event):
        log.debug(f"on_window_key_press: {window.get_title()} {event}")

        # debug leaking memory
        if Gdk.keyval_name(event.keyval) == "F12":
            from pychess.System.debug import print_obj_referrers, print_muppy_sumary

            if event.get_state() & Gdk.ModifierType.SHIFT_MASK:
                print_muppy_sumary()
            else:
                print_obj_referrers()

        # Tabbing related shortcuts
        persp = perspective_manager.get_perspective("games")
        if not persp.getheadbook():
            pagecount = 0
        else:
            pagecount = persp.getheadbook().get_n_pages()
        if pagecount > 1:
            if event.get_state() & Gdk.ModifierType.CONTROL_MASK:
                page_num = persp.getheadbook().get_current_page()
                # Move selected
                if event.get_state() & Gdk.ModifierType.SHIFT_MASK:
                    child = persp.getheadbook().get_nth_page(page_num)
                    if event.keyval == Gdk.KEY_Page_Up:
                        persp.getheadbook().reorder_child(
                            child, (page_num - 1) % pagecount
                        )
                        return True
                    elif event.keyval == Gdk.KEY_Page_Down:
                        persp.getheadbook().reorder_child(
                            child, (page_num + 1) % pagecount
                        )
                        return True
                # Change selected
                else:
                    if event.keyval == Gdk.KEY_Page_Up:
                        persp.getheadbook().set_current_page((page_num - 1) % pagecount)
                        return True
                    elif event.keyval == Gdk.KEY_Page_Down:
                        persp.getheadbook().set_current_page((page_num + 1) % pagecount)
                        return True

        gmwidg = persp.cur_gmwidg()
        if gmwidg is not None:
            # Let default handler work if typing inside entry widgets
            current_focused_widget = gamewidget.getWidgets()["main_window"].get_focus()
            if current_focused_widget is not None and isinstance(
                current_focused_widget, Gtk.Entry
            ):
                return False

            # Prevent moving in game while lesson not finished
            if gmwidg.gamemodel.lesson_game:
                return

            # Navigate on boardview with arrow keys
            if event.keyval in leftkeys:
                if event.get_state() & Gdk.ModifierType.CONTROL_MASK:
                    gmwidg.board.view.backToMainLine()
                    return True
                else:
                    gmwidg.board.view.showPrev()
                    return True
            elif event.keyval in rightkeys:
                gmwidg.board.view.showNext()
                return True
            elif event.keyval in upkeys:
                gmwidg.board.view.showPrev(step=2)
                return True
            elif event.keyval in downkeys:
                gmwidg.board.view.showNext(step=2)
                return True
            elif event.keyval in homekeys:
                gmwidg.board.view.showFirst()
                return True
            elif event.keyval in endkeys:
                gmwidg.board.view.showLast()
                return True

            if (
                (not event.get_state() & Gdk.ModifierType.CONTROL_MASK)
                and (not event.get_state() & Gdk.ModifierType.MOD1_MASK)
                and (event.keyval != Gdk.KEY_Escape)
                and (event.keyval not in functionkeys)
            ):
                # Enter moves with keyboard
                board_control = gmwidg.board
                keyname = Gdk.keyval_name(event.keyval)
                board_control.key_pressed(keyname)
                print(board_control.keybuffer)
                return True

            return False

    def on_recent_game_activated(self, uri):
        if isinstance(uri, str):
            path = url2pathname(uri)
            recent_manager.add_item("file:" + pathname2url(path))

    # Drag 'n' Drop

    def on_drag_received(
        self, widget, context, x, y, selection, target_type, timestamp
    ):
        if target_type == TARGET_TYPE_URI_LIST:
            NOTPROC, PARTIAL, FULL = range(3)
            status = NOTPROC
            uris = selection.get_uris()
            for uri in uris:
                fn, fext = os.path.splitext(uri.lower())
                b = False

                # Chess position
                if fext == ".fen":
                    b = newGameDialog.loadFileAndRun(uri)

                # Shortcut
                elif fext in [".url", ".desktop"]:
                    # Preconf
                    if fext == ".url":
                        sectname = "InternetShortcut"
                        typeok = True
                    elif fext == ".desktop":
                        sectname = "Desktop Entry"
                        typeok = False
                    else:
                        assert False

                    # Read the shortcut
                    filename = splitUri(uri)[1]
                    with open(filename) as file:
                        content = file.read()
                    lines = content.replace("\r", "").split("\n")

                    # Extract the link
                    section = False
                    link = ""
                    for item in lines:
                        # Header
                        if item.startswith("["):
                            if section:
                                break
                            section = item.startswith("[%s]" % sectname)
                        if not section:
                            continue

                        # Item
                        if item.startswith("URL="):
                            link = item[4:]
                        if item.startswith("Type=Link") and fext == ".desktop":
                            typeok = True

                    # Load the link
                    if typeok and link != "":
                        asyncio.create_task(get_internet_game(link))
                        b = True

                # Database
                else:
                    perspective = perspective_manager.get_perspective("database")
                    perspective.open_chessfile(uri)
                    b = True

                # Update the global status
                if b:
                    if status == NOTPROC:
                        status = FULL
                else:
                    if status != NOTPROC:
                        status = PARTIAL

            # Feedback about the load
            msg = ""
            if status == NOTPROC:
                msg = _("All the links failed to fetch a relevant chess content.")
                msgtype = Gtk.MessageType.ERROR
            elif status == PARTIAL:
                msg = _("Some links were invalid.")
                msgtype = Gtk.MessageType.WARNING
            if msg != "":
                dlg = Gtk.MessageDialog(
                    mainwindow(),
                    type=msgtype,
                    buttons=Gtk.ButtonsType.OK,
                    message_format=msg,
                )
                dlg.run()
                dlg.destroy()

    # Game Menu

    def on_new_game1_activate(self, widget):
        newGameDialog.NewGameMode.run()

    def on_set_up_position_activate(self, widget):
        rotate_menu = gamewidget.getWidgets()["rotate_board1"]
        rotate_menu.set_sensitive(True)
        persp = perspective_manager.get_perspective("games")
        gmwidg = persp.cur_gmwidg()
        if gmwidg is not None:
            ply = gmwidg.board.view.shown
            variation = gmwidg.board.view.shown_variation_idx
            board = gmwidg.gamemodel.getBoardAtPly(ply, variation)
            fen = board.asFen()
            variant = board.variant
        else:
            fen = None
            variant = NORMALCHESS
        newGameDialog.SetupPositionExtension.run(fen, variant)

    def on_enter_game_notation_activate(self, widget):
        newGameDialog.EnterNotationExtension.run()

    def on_enter_game_clipboard_activate(self, widget):
        ok = False
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        text = clipboard.wait_for_text()
        if text is not None:
            text = text.strip()
            if text != "":
                try:
                    pgn = PGNFile(StringIO(text))
                    pgn.loadToModel(None)
                    newGameDialog.loadPgnAndRun(text)
                    ok = True
                except Exception:
                    pass
        if not ok:
            dialog = Gtk.MessageDialog(
                mainwindow(), type=Gtk.MessageType.ERROR, buttons=Gtk.ButtonsType.OK
            )
            dialog.set_markup(_("The clipboard contains no relevant chess data."))
            dialog.run()
            dialog.destroy()

    def on_play_internet_chess_activate(self, widget):
        ICLogon.run()

    def on_load_game1_activate(self, widget):
        dialog = get_open_dialog()

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            filenames = dialog.get_filenames()
        else:
            filenames = None

        dialog.destroy()

        if filenames is not None:
            for filename in filenames:
                if filename.lower().endswith(".fen"):
                    newGameDialog.loadFileAndRun(filename)
                else:
                    perspective = perspective_manager.get_perspective("database")
                    perspective.open_chessfile(filename)

    def on_remote_game_activate(self, widget):
        url = getUserTextDialog(
            mainwindow(), _("Load a remote game"), _("Paste the link to download:")
        )
        asyncio.create_task(get_internet_game(url))

    def on_save_game1_activate(self, widget):
        perspective = perspective_manager.get_perspective("games")
        gmwidg = perspective.cur_gmwidg()
        position = gmwidg.board.view.shown
        perspective.saveGame(gmwidg.gamemodel, position)

    def on_save_game_as1_activate(self, widget):
        perspective = perspective_manager.get_perspective("games")
        gmwidg = perspective.cur_gmwidg()
        position = gmwidg.board.view.shown
        perspective.saveGameAs(gmwidg.gamemodel, position)

    def on_export_position_activate(self, widget):
        perspective = perspective_manager.get_perspective("games")
        gmwidg = perspective.cur_gmwidg()
        position = gmwidg.board.view.shown
        perspective = perspective_manager.get_perspective("games")
        perspective.saveGameAs(gmwidg.gamemodel, position, export=True)

    def on_share_game_activate(self, widget):
        perspective = perspective_manager.get_perspective("games")
        gmwidg = perspective.cur_gmwidg()
        chesspastebin.paste(gmwidg.gamemodel)

    def on_properties1_activate(self, widget):
        gameinfoDialog.run(gamewidget.getWidgets())

    def on_analyze_game_activate(self, widget):
        analyze_game_dialog = AnalyzeGameDialog()
        analyze_game_dialog.run()

    def on_player_rating1_activate(self, widget):
        playerinfoDialog.run(gamewidget.getWidgets())

    def on_close1_activate(self, widget):
        persp = perspective_manager.current_perspective
        if persp.name == "games":
            gmwidg = persp.cur_gmwidg()
            persp.closeGame(gmwidg)
        elif persp.name == "database":
            persp.close()

    def on_quit1_activate(self, widget, *args):
        perspective = perspective_manager.get_perspective("games")
        if isinstance(widget, Gdk.Event):
            if len(perspective.gamewidgets) == 1 and conf.get("hideTabs"):
                gmwidg = perspective.cur_gmwidg()
                perspective.closeGame(gmwidg, gmwidg.gamemodel)
                return True
            elif len(perspective.gamewidgets) >= 1 and conf.get("closeAll"):
                perspective.closeAllGames(perspective.gamewidgets)
                return True
        if perspective.closeAllGames(perspective.gamewidgets) in (
            Gtk.ResponseType.OK,
            Gtk.ResponseType.YES,
        ):
            ICLogon.stop()
            asyncio.get_running_loop().stop()
            self.app.quit()
        else:
            return True

    # View Menu

    def on_rotate_board1_activate(self, widget):
        board_control = newGameDialog.SetupPositionExtension.board_control
        persp = perspective_manager.get_perspective("games")
        if board_control is not None and board_control.view.is_visible():
            view = newGameDialog.SetupPositionExtension.board_control.view
        elif persp.cur_gmwidg() is not None:
            view = persp.cur_gmwidg().board.view
        else:
            return
        if view.rotation:
            view.rotation = 0
        else:
            view.rotation = math.pi

    def on_fullscreen1_activate(self, widget):
        gamewidget.getWidgets()["main_window"].fullscreen()
        gamewidget.getWidgets()["fullscreen1"].hide()
        gamewidget.getWidgets()["leave_fullscreen1"].show()

    def on_leave_fullscreen1_activate(self, widget):
        gamewidget.getWidgets()["main_window"].unfullscreen()
        gamewidget.getWidgets()["leave_fullscreen1"].hide()
        gamewidget.getWidgets()["fullscreen1"].show()

    def on_about1_activate(self, widget):
        about_dialog = gamewidget.getWidgets()["aboutdialog1"]
        response = about_dialog.run()
        if (
            response == Gtk.ResponseType.DELETE_EVENT
            or response == Gtk.ResponseType.CANCEL
        ):
            gamewidget.getWidgets()["aboutdialog1"].hide()

    def on_log_viewer1_activate(self, widget):
        from pychess.widgets import LogDialog

        if widget.get_active():
            LogDialog.show()
        else:
            LogDialog.hide()

    def on_show_sidepanels_activate(self, widget):
        perspective = perspective_manager.get_perspective("games")
        if perspective is not None:
            perspective.zoomToBoard(not widget.get_active())

    def on_hint_mode_activate(self, widget):
        perspective = perspective_manager.get_perspective("games")
        if perspective is None:
            return

        for gmwidg in perspective.gamewidgets:
            if gmwidg.isInFront():
                try:
                    analyzer = gmwidg.gamemodel.spectators[HINT]
                except KeyError:
                    continue
                if widget.get_active():
                    gmwidg.show_arrow(analyzer, HINT)
                else:
                    gmwidg.hide_arrow(analyzer, HINT)

    def on_spy_mode_activate(self, widget):
        perspective = perspective_manager.get_perspective("games")
        if perspective is None:
            return
        for gmwidg in perspective.gamewidgets:
            if gmwidg.isInFront():
                try:
                    analyzer = gmwidg.gamemodel.spectators[HINT]
                except KeyError:
                    continue
                if widget.get_active():
                    gmwidg.show_arrow(analyzer, SPY)
                else:
                    gmwidg.hide_arrow(analyzer, SPY)

    # Edit menu

    def on_copy_pgn_activate(self, widget):
        persp = perspective_manager.get_perspective("games")
        if perspective_manager.current_perspective == persp:
            persp.cur_gmwidg().board.view.copy_pgn()
            return

        persp = perspective_manager.get_perspective("database")
        if perspective_manager.current_perspective == persp:
            if persp.preview_panel is not None:
                persp.preview_panel.boardview.copy_pgn()

    def on_copy_fen_activate(self, widget):
        persp = perspective_manager.get_perspective("games")
        if perspective_manager.current_perspective == persp:
            persp.cur_gmwidg().board.view.copy_fen()
            return

        persp = perspective_manager.get_perspective("database")
        if perspective_manager.current_perspective == persp:
            if persp.preview_panel is not None:
                persp.preview_panel.boardview.copy_fen()

    def on_manage_engines_activate(self, widget):
        enginesDialog.run(gamewidget.getWidgets())

    def on_download_externals_activate(self, widget):
        externals_dialog = ExternalsDialog()
        externals_dialog.show()

    def on_preferences_activate(self, widget):
        preferencesDialog.run(gamewidget.getWidgets())

    # Database menu

    def on_new_database1_activate(self, widget):
        perspective = perspective_manager.get_perspective("database")
        perspective.create_database()

    def on_import_chessfile_activate(self, widget):
        perspective = perspective_manager.get_perspective("database")
        perspective.on_import_clicked(widget)

    def on_database_save_as_activate(self, widget):
        perspective = perspective_manager.get_perspective("database")
        perspective.on_save_as_clicked(widget)

    def on_create_book_activate(self, widget):
        perspective = perspective_manager.get_perspective("database")
        perspective.create_book()

    def on_import_endgame_nl_activate(self, widget):
        perspective = perspective_manager.get_perspective("database")
        perspective.on_import_endgame_nl()

    def on_import_twic_activate(self, widget):
        perspective = perspective_manager.get_perspective("database")
        perspective.on_import_twic()

    def on_update_players_activate(self, widget):
        perspective = perspective_manager.get_perspective("database")
        perspective.on_update_players()

    # Help menu

    def on_about_chess1_activate(self, widget):
        webbrowser.open(_("http://en.wikipedia.org/wiki/Chess"))

    def on_how_to_play1_activate(self, widget):
        webbrowser.open(_("http://en.wikipedia.org/wiki/Rules_of_chess"))

    def on_TipOfTheDayMenuItem_activate(self, widget):
        tip_of_the_day = TipOfTheDay()
        tip_of_the_day.show()

    def translate_this_application_activate(self, widget):
        webbrowser.open("https://www.transifex.com/projects/p/pychess/")

    def report_issue_activate(self, widget):
        webbrowser.open("https://github.com/pychess/pychess/issues")


class PyChess(Gtk.Application):
    def __init__(
        self,
        log_viewer,
        purge_recent,
        chess_file,
        ics_host,
        ics_port,
        splash,
        version_check,
    ):
        Gtk.Application.__init__(
            self, application_id="org.pychess", flags=Gio.ApplicationFlags.NON_UNIQUE
        )

        if ics_host:
            ICLogon.host = ics_host
        if ics_port:
            ICLogon.port = ics_port

        self.log_viewer = log_viewer
        self.purge_recent = purge_recent
        self.chess_file = chess_file
        self.window = None
        self.splash = splash
        self.version_check = version_check

    def do_startup(self):
        Gtk.Application.do_startup(self)

        if self.purge_recent:
            items = recent_manager.get_items()
            for item in items:
                uri = item.get_uri()
                if item.get_application_info("pychess"):
                    recent_manager.remove_item(uri)

        self.git_rev = ""

        self.initGlade(self.log_viewer)
        self.addPerspectives()
        self.handleArgs(self.chess_file)
        if self.version_check:
            asyncio.create_task(checkversion())

        self.loaded_cids = {}
        self.saved_cids = {}
        self.terminated_cids = {}

        log.info(f"PyChess {VERSION_NAME} {VERSION} git {self.git_rev}")
        log.info("Command line args: '%s'" % self.chess_file)
        log.info("Platform: %s" % platform.platform())
        log.info("Python version: %s.%s.%s" % sys.version_info[0:3])
        log.info("Pyglib version: %s.%s.%s" % GLib.pyglib_version)
        log.info(
            "Gtk version: {}.{}.{}".format(
                Gtk.get_major_version(),
                Gtk.get_minor_version(),
                Gtk.get_micro_version(),
            )
        )

    async def print_tasks(self):
        while True:
            print(datetime.datetime.now().time())
            loop = asyncio.get_event_loop()
            tasks = asyncio.Task.all_tasks(loop)
            for task in tasks:
                print(task)
            print("------------")
            await asyncio.sleep(10)

    def do_activate(self):
        # asyncio.create_task(self.print_tasks())
        self.add_window(self.window)
        self.window.show_all()
        gamewidget.getWidgets()["player_rating1"].hide()
        gamewidget.getWidgets()["leave_fullscreen1"].hide()

        # Externals download dialog
        if not conf.get("dont_show_externals_at_startup"):
            externals_dialog = ExternalsDialog()
            externals_dialog.show()

        # Tip of the day dialog
        if conf.get("show_tip_at_startup"):
            tip_of_the_day = TipOfTheDay()
            tip_of_the_day.show()

        preferencesDialog.run(gamewidget.getWidgets())

        def on_all_engine_discovered(discoverer):
            engine = discoverer.getEngineByName(discoverer.getEngineLearn())
            if engine is None:
                engine = discoverer.getEngineN(-1)
            default_engine = engine.get("md5")
            conf.set("ana_combobox", default_engine)
            conf.set("inv_ana_combobox", default_engine)

        # Try to set conf analyzer engine on very first start of pychess
        if conf.get("ana_combobox") == 0:
            discoverer.connect_after("all_engines_discovered", on_all_engine_discovered)

        dd = DiscovererDialog(discoverer)
        self.dd_task = asyncio.create_task(dd.start())

        style_ctxt = gamewidget.getWidgets()["main_window"].get_style_context()
        LIGHT = hexcol(style_ctxt.lookup_color("p_light_color")[1])
        DARK = hexcol(style_ctxt.lookup_color("p_dark_color")[1])
        conf.DEFAULTS["General"]["lightcolour"] = LIGHT
        conf.DEFAULTS["General"]["darkcolour"] = DARK

        self.splash.destroy()

    def on_gmwidg_created(self, gamehandler, gmwidg):
        log.debug("GladeHandlers.on_gmwidg_created: starting")
        # Bring playing window to the front
        gamewidget.getWidgets()["main_window"].present()

        self.loaded_cids[gmwidg.gamemodel] = gmwidg.gamemodel.connect(
            "game_loaded", self.update_recent
        )
        self.saved_cids[gmwidg.gamemodel] = gmwidg.gamemodel.connect(
            "game_saved", self.update_recent
        )
        self.terminated_cids[gmwidg.gamemodel] = gmwidg.gamemodel.connect(
            "game_terminated", self.on_terminated
        )

        log.debug("GladeHandlers.on_gmwidg_created: returning")

    def on_chessfile_opened(self, persp, chessfile):
        self.update_recent(None, chessfile.path)

    def on_terminated(self, gamemodel):
        if gamemodel.handler_is_connected(self.loaded_cids[gamemodel]):
            gamemodel.disconnect(self.loaded_cids[gamemodel])
            del self.loaded_cids[gamemodel]
        if gamemodel.handler_is_connected(self.saved_cids[gamemodel]):
            gamemodel.disconnect(self.saved_cids[gamemodel])
            del self.saved_cids[gamemodel]
        if gamemodel.handler_is_connected(self.terminated_cids[gamemodel]):
            gamemodel.disconnect(self.terminated_cids[gamemodel])
            del self.terminated_cids[gamemodel]

    def update_recent(self, gamemodel, uri):
        if isinstance(uri, str):
            path = url2pathname(uri)
            recent_manager.add_item("file:" + pathname2url(path))

    def initGlade(self, log_viewer):
        # Init glade and the 'GladeHandlers'
        self.widgets = widgets = uistuff.GladeWidgets("PyChess.glade")
        self.glade_handlers = GladeHandlers(self)
        widgets.getGlade().connect_signals(self.glade_handlers)
        self.window = widgets["main_window"]

        # new_game_tasker, internet_game_tasker = NewGameTasker(
        # ), InternetGameTasker()
        # tasker.packTaskers(new_game_tasker, internet_game_tasker)
        # widgets["Background"].add(tasker)

        # Redirect widgets
        gamewidget.setWidgets(widgets)

        # The only menuitems that need special initing
        for widget in ("hint_mode", "spy_mode"):
            widgets[widget].set_sensitive(False)

        uistuff.keep(widgets["hint_mode"], "hint_mode")
        uistuff.keep(widgets["spy_mode"], "spy_mode")
        uistuff.keep(widgets["show_sidepanels"], "show_sidepanels")
        uistuff.keep(widgets["auto_call_flag"], "autoCallFlag")

        # Show main window and init d'n'd
        widgets["main_window"].set_title("%s - PyChess" % _("Welcome"))
        widgets["main_window"].connect(
            "delete-event", self.glade_handlers.on_quit1_activate
        )
        widgets["main_window"].connect(
            "key-press-event", self.glade_handlers.on_window_key_press
        )

        uistuff.keepWindowSize(
            "main", widgets["main_window"], None, uistuff.POSITION_GOLDEN
        )

        # To get drag in the whole window, we add it to the menu and the
        # background. If it can be gotten to work, the drag_dest_set_proxy
        # method is very interesting.
        widgets["menubar1"].drag_dest_set(Gtk.DestDefaults.ALL, DND_LIST, DRAG_ACTION)
        widgets["box2"].drag_dest_set(Gtk.DestDefaults.ALL, DND_LIST, DRAG_ACTION)
        widgets["perspectives_notebook"].drag_dest_set(
            Gtk.DestDefaults.ALL, DND_LIST, DRAG_ACTION
        )

        # Init 'minor' dialogs

        # Log dialog
        if log_viewer:
            from pychess.widgets import LogDialog

            LogDialog.add_destroy_notify(lambda: widgets["log_viewer1"].set_active(0))
        else:
            widgets["log_viewer1"].set_property("sensitive", False)

        # About dialog
        self.aboutdialog = widgets["aboutdialog1"]
        self.aboutdialog.set_program_name(NAME)
        self.aboutdialog.set_copyright("Copyright © 2006-2024")
        self.aboutdialog.set_version(VERSION_NAME + " " + VERSION)
        if isgit():
            try:
                self.git_rev = (
                    subprocess.check_output(["git", "describe", "--tags"])
                    .decode()
                    .strip()
                )
                self.aboutdialog.set_version(f"{VERSION_NAME} Git {self.git_rev}")
            except subprocess.CalledProcessError:
                pass
        self.aboutdialog.set_comments(self.aboutdialog.get_comments())

        with open(prefix.addDataPrefix("ARTISTS"), encoding="utf-8") as f:
            self.aboutdialog.set_artists(f.read().splitlines())
        with open(prefix.addDataPrefix("AUTHORS"), encoding="utf-8") as f:
            self.aboutdialog.set_authors(f.read().splitlines())
        with open(prefix.addDataPrefix("DOCUMENTERS"), encoding="utf-8") as f:
            self.aboutdialog.set_documenters(f.read().splitlines())
        with open(prefix.addDataPrefix("TRANSLATORS"), encoding="utf-8") as f:
            self.aboutdialog.set_translator_credits(f.read())
        with open(prefix.addDataPrefix("LICENSE"), encoding="utf-8") as f:
            self.aboutdialog.set_license(f.read())

        widgets["load_recent_game1"].set_submenu(recent_menu)

        if conf.get("autoLogin"):
            internet_game_tasker.connectClicked(None)

    def website(self, clb, link):
        webbrowser.open(link)

    def addPerspectives(self):
        perspective_manager.set_widgets(self.widgets)
        for persp in (Welcome, Games, FICS, Database, Learn):
            perspective = persp()
            perspective_manager.add_perspective(perspective)
            perspective.create_toolbuttons()

            if persp == Database:
                perspective.connect("chessfile_opened", self.on_chessfile_opened)
            elif persp == Games:
                perspective.connect("gmwidg_created", self.on_gmwidg_created)

        newimage = get_pixbuf("glade/new24.png")
        new_button = Gtk.ToolButton.new(Gtk.Image.new_from_pixbuf(newimage))
        new_button.set_tooltip_text(_("New Game"))
        new_button.connect("clicked", self.glade_handlers.on_new_game1_activate)
        perspective_manager.toolbar.insert(new_button, 0)

        openimage = get_pixbuf("glade/open24.png")
        open_button = Gtk.ToolButton.new(Gtk.Image.new_from_pixbuf(openimage))
        open_button.set_tooltip_text(_("Open Game"))
        open_button.connect("clicked", self.glade_handlers.on_load_game1_activate)
        perspective_manager.toolbar.insert(open_button, 1)

    def handleArgs(self, chess_file):
        if chess_file:

            def do(discoverer):
                perspective = perspective_manager.get_perspective("database")
                perspective.open_chessfile(chess_file)
                self.dd_task.cancel()

            discoverer.connect_after("all_engines_discovered", do)



from .bitboard import bitPosArray, iterBits, clearBit, firstBit
from .attack import isAttacked, pinnedOnKing, getAttacks
from .ldata import (
    fromToRay,
    moveArray,
    directions,
    fileBits,
    rankBits,
    ray45,
    attack45,
    ray135,
    attack135,
    ray90,
    attack90,
    ray00,
    attack00,
    FILE,
    rays,
)
from pychess.Utils.const import (
    EMPTY,
    PAWN,
    QUEEN,
    KNIGHT,
    BISHOP,
    ROOK,
    KING,
    HAWK,
    ELEPHANT,
    WHITE,
    BLACK,
    SITTUYINCHESS,
    FISCHERRANDOMCHESS,
    SUICIDECHESS,
    GIVEAWAYCHESS,
    CAMBODIANCHESS,
    ATOMICCHESS,
    WILDCASTLECHESS,
    WILDCASTLESHUFFLECHESS,
    CRAZYHOUSECHESS,
    ASEAN_VARIANTS,
    HORDECHESS,
    PLACEMENTCHESS,
    SCHESS,
    LIGHTBRIGADECHESS,
    BPAWN,
    sliders,
    A8,
    A6,
    G6,
    F6,
    H1,
    C3,
    B2,
    B3,
    A3,
    D6,
    D8,
    E3,
    E1,
    E8,
    C7,
    F2,
    D1,
    E6,
    H3,
    D3,
    H2,
    G7,
    H6,
    H7,
    ASEAN_QUEEN,
    ASEAN_BBISHOP,
    ASEAN_WBISHOP,
    NORMAL_MOVE,
    QUEEN_CASTLE,
    KING_CASTLE,
    ENPASSANT,
    KNIGHT_PROMOTION,
    BISHOP_PROMOTION,
    ROOK_PROMOTION,
    QUEEN_PROMOTION,
    KING_PROMOTION,
    DROP_VARIANTS,
    DROP,
    B_OOO,
    B_OO,
    W_OOO,
    W_OO,
    HAWK_PROMOTION,
    ELEPHANT_PROMOTION,
    HAWK_GATE,
    ELEPHANT_GATE,
    HAWK_GATE_AT_ROOK,
    ELEPHANT_GATE_AT_ROOK,
)

# The format of a move is as follows - from left:
# 4 bits:  Descriping the type of the move
# 6 bits:  cord to move from
# 6 bits:  cord to move to

shiftedFromCords = []
for i in range(64):
    shiftedFromCords.append(i << 6)

shiftedFlags = []
for i in (
    NORMAL_MOVE,
    QUEEN_CASTLE,
    KING_CASTLE,
    ENPASSANT,
    KNIGHT_PROMOTION,
    BISHOP_PROMOTION,
    ROOK_PROMOTION,
    QUEEN_PROMOTION,
    KING_PROMOTION,
    HAWK_PROMOTION,
    ELEPHANT_PROMOTION,
    HAWK_GATE,
    ELEPHANT_GATE,
    HAWK_GATE_AT_ROOK,
    ELEPHANT_GATE_AT_ROOK,
    DROP,
):
    shiftedFlags.append(i << 12)


def newMove(fromcord, tocord, flag=NORMAL_MOVE):
    return shiftedFlags[flag] + shiftedFromCords[fromcord] + tocord


# Generate all moves


def genCastles(board):
    def generateOne(color, side, king_after, rook_after):
        if side == 0:
            castle = QUEEN_CASTLE
        else:
            castle = KING_CASTLE
        king = board.ini_kings[color]
        rook = board.ini_rooks[color][side]
        blocker = clearBit(clearBit(board.blocker, king), rook)
        stepover = fromToRay[king][king_after] | fromToRay[rook][rook_after]
        if not stepover & blocker:
            for cord in range(min(king, king_after), max(king, king_after) + 1):
                if isAttacked(board, cord, 1 - color):
                    return
            if FILE(king) == 3 and board.variant in (
                WILDCASTLECHESS,
                WILDCASTLESHUFFLECHESS,
            ):
                castle = QUEEN_CASTLE if castle == KING_CASTLE else KING_CASTLE
            if board.variant == FISCHERRANDOMCHESS:
                yield newMove(king, rook, castle)
            else:
                yield newMove(king, king_after, castle)
                if board.variant == SCHESS:
                    holding = board.holding[color]
                    if holding[HAWK] == 0 and holding[ELEPHANT] == 0:
                        return
                    for king in iterBits(
                        board.boards[board.color][KING] & board.virgin[board.color]
                    ):
                        if holding[HAWK] > 0:
                            yield newMove(king, king_after, HAWK_GATE)
                            yield newMove(rook, king, HAWK_GATE_AT_ROOK)
                        if holding[ELEPHANT] > 0:
                            yield newMove(king, king_after, ELEPHANT_GATE)
                            yield newMove(rook, king, ELEPHANT_GATE_AT_ROOK)

    king = board.ini_kings[board.color]
    wildcastle = FILE(king) == 3 and board.variant in (
        WILDCASTLECHESS,
        WILDCASTLESHUFFLECHESS,
    )
    if board.color == WHITE:
        if board.castling & W_OO:
            side = 0 if wildcastle else 1
            yield from generateOne(
                WHITE, side, board.fin_kings[WHITE][side], board.fin_rooks[WHITE][side]
            )

        if board.castling & W_OOO:
            side = 1 if wildcastle else 0
            yield from generateOne(
                WHITE, side, board.fin_kings[WHITE][side], board.fin_rooks[WHITE][side]
            )
    else:
        if board.castling & B_OO:
            side = 0 if wildcastle else 1
            yield from generateOne(
                BLACK, side, board.fin_kings[BLACK][side], board.fin_rooks[BLACK][side]
            )

        if board.castling & B_OOO:
            side = 1 if wildcastle else 0
            yield from generateOne(
                BLACK, side, board.fin_kings[BLACK][side], board.fin_rooks[BLACK][side]
            )


def genPieceMoves(board, piece, tcord):
    """ "
    Used by parseSAN only to accelerate it a bit
    """
    moves = set()
    friends = board.friends[board.color]
    notfriends = ~friends
    if piece == KNIGHT:
        knights = board.boards[board.color][KNIGHT]
        knightMoves = moveArray[KNIGHT]
        for fcord in iterBits(knights):
            if tcord in iterBits(knightMoves[fcord] & notfriends):
                moves.add(newMove(fcord, tcord))
        return moves

    if piece == BISHOP:
        bishops = board.boards[board.color][BISHOP]
        if board.variant in ASEAN_VARIANTS:
            bishopMoves = moveArray[
                ASEAN_WBISHOP if board.color == WHITE else ASEAN_BBISHOP
            ]
            for fcord in iterBits(bishops):
                if tcord in iterBits(bishopMoves[fcord] & notfriends):
                    moves.add(newMove(fcord, tcord))
            return moves
        else:
            blocker = board.blocker
            for fcord in iterBits(bishops):
                try:
                    attackBoard = (
                        attack45[fcord][ray45[fcord] & blocker]
                        | attack135[fcord][ray135[fcord] & blocker]
                    )
                except KeyError:
                    attackBoard = 0
                if tcord in iterBits(attackBoard & notfriends):
                    moves.add(newMove(fcord, tcord))
            return moves

    if piece == ROOK:
        blocker = board.blocker
        rooks = board.boards[board.color][ROOK]
        for fcord in iterBits(rooks):
            try:
                attackBoard = (
                    attack00[fcord][ray00[fcord] & blocker]
                    | attack90[fcord][ray90[fcord] & blocker]
                )
            except KeyError:
                attackBoard = 0
            if tcord in iterBits(attackBoard & notfriends):
                moves.add(newMove(fcord, tcord))
        return moves

    if piece == QUEEN:
        queens = board.boards[board.color][QUEEN]
        if board.variant in ASEAN_VARIANTS:
            queenMoves = moveArray[ASEAN_QUEEN]
            for fcord in iterBits(queens):
                if tcord in iterBits(queenMoves[fcord] & notfriends):
                    moves.add(newMove(fcord, tcord))
            # Cambodian extra first move
            if board.variant == CAMBODIANCHESS:
                if board.is_first_move[QUEEN][board.color]:
                    if board.color == WHITE:
                        if not board.arBoard[E3]:
                            moves.add(newMove(E1, E3))
                    else:
                        if not board.arBoard[D6]:
                            moves.add(newMove(D8, D6))
            return moves
        else:
            blocker = board.blocker
            for fcord in iterBits(queens):
                try:
                    attackBoard = (
                        attack45[fcord][ray45[fcord] & blocker]
                        | attack135[fcord][ray135[fcord] & blocker]
                    )
                except KeyError:
                    attackBoard = 0
                if tcord in iterBits(attackBoard & notfriends):
                    moves.add(newMove(fcord, tcord))

                try:
                    attackBoard = (
                        attack00[fcord][ray00[fcord] & blocker]
                        | attack90[fcord][ray90[fcord] & blocker]
                    )
                except KeyError:
                    attackBoard = 0
                if tcord in iterBits(attackBoard & notfriends):
                    moves.add(newMove(fcord, tcord))
            return moves

    if (
        board.variant == SUICIDECHESS or board.variant == GIVEAWAYCHESS
    ) and piece == KING:
        kings = board.boards[board.color][KING]
        if kings:
            kingMoves = moveArray[KING]
            for fcord in iterBits(kings):
                for tc in iterBits(kingMoves[fcord] & notfriends):
                    if tc == tcord:
                        moves.add(newMove(fcord, tcord))
            return moves

    return moves


def gen_sittuyin_promotions(board):
    from pychess.Variants import variants

    blocker = board.blocker
    notblocker = ~blocker

    pawns = board.boards[board.color][PAWN]

    queenMoves = moveArray[ASEAN_QUEEN]

    def willDirectAttack(board, move, cord):
        board_clone = board.clone()
        board_clone.applyMove(move)
        return board.friends[1 - board.color] & moveArray[ASEAN_QUEEN][cord]

    promotion_zone = variants[SITTUYINCHESS].PROMOTION_ZONE[board.color]
    for cord in iterBits(pawns):
        if board.pieceCount[board.color][PAWN] == 1 or cord in promotion_zone:
            # in place promotions
            move = newMove(cord, cord, QUEEN_PROMOTION)
            if not board.willGiveCheck(move) and not willDirectAttack(
                board, move, cord
            ):
                yield move

            # queen move promotion
            for c in iterBits(queenMoves[cord] & notblocker):
                move = newMove(cord, c, QUEEN_PROMOTION)
                if not board.willGiveCheck(move) and not willDirectAttack(
                    board, move, c
                ):
                    yield move


def genAllMoves(board, drops=True):
    from pychess.Variants import variants

    if drops and board.variant in DROP_VARIANTS and board.variant != SCHESS:
        yield from genDrops(board)

    # In sittuyin you have to place your pieces before any real move
    if board.variant == SITTUYINCHESS or board.variant == PLACEMENTCHESS:
        if board.plyCount < 16:
            return

    blocker = board.blocker
    notblocker = ~blocker
    enpassant = board.enpassant

    friends = board.friends[board.color]
    notfriends = ~friends
    enemies = board.friends[1 - board.color]

    pawns = board.boards[board.color][PAWN]
    knights = board.boards[board.color][KNIGHT]
    bishops = board.boards[board.color][BISHOP]
    rooks = board.boards[board.color][ROOK]
    queens = board.boards[board.color][QUEEN]
    kings = board.boards[board.color][KING]
    hawks = board.boards[board.color][HAWK]
    elephants = board.boards[board.color][ELEPHANT]

    PROMOTIONS = variants[board.variant].PROMOTIONS
    # In sittuyin only one queen allowed to exist any time per side
    if board.variant == SITTUYINCHESS and queens:
        PROMOTIONS = (NORMAL_MOVE,)
    elif board.variant == LIGHTBRIGADECHESS:
        PROMOTIONS = (QUEEN_PROMOTION,) if board.color == WHITE else (KNIGHT_PROMOTION,)

    # Knights, Hawks, Elephants
    knightMoves = moveArray[KNIGHT]
    for cord in iterBits(knights | hawks | elephants):
        for c in iterBits(knightMoves[cord] & notfriends):
            yield newMove(cord, c)

    # knight moves with gating
    holding = board.holding[board.color]
    if board.variant == SCHESS and (holding[HAWK] > 0 or holding[ELEPHANT] > 0):
        for cord in iterBits(knights & board.virgin[board.color]):
            for c in iterBits(knightMoves[cord] & notfriends):
                if holding[HAWK] > 0:
                    yield newMove(cord, c, HAWK_GATE)
                if holding[ELEPHANT] > 0:
                    yield newMove(cord, c, ELEPHANT_GATE)

    # King
    if kings:
        kingMoves = moveArray[KING]
        # cord = firstBit(kings)
        for cord in iterBits(kings):
            for c in iterBits(kingMoves[cord] & notfriends):
                if board.variant == ATOMICCHESS:
                    if not board.arBoard[c]:
                        yield newMove(cord, c)
                else:
                    yield newMove(cord, c)
                    if board.variant == SCHESS and (
                        holding[HAWK] > 0 or holding[ELEPHANT] > 0
                    ):
                        for cord in iterBits(kings & board.virgin[board.color]):
                            if holding[HAWK] > 0:
                                yield newMove(cord, c, HAWK_GATE)
                            if holding[ELEPHANT] > 0:
                                yield newMove(cord, c, ELEPHANT_GATE)

    if board.variant in ASEAN_VARIANTS:
        # Rooks
        for cord in iterBits(rooks):
            try:
                attackBoard = (
                    attack00[cord][ray00[cord] & blocker]
                    | attack90[cord][ray90[cord] & blocker]
                )
            except KeyError:
                attackBoard = 0
            for c in iterBits(attackBoard & notfriends):
                yield newMove(cord, c)

        # Queens
        queenMoves = moveArray[ASEAN_QUEEN]
        for cord in iterBits(queens):
            for c in iterBits(queenMoves[cord] & notfriends):
                yield newMove(cord, c)

        # Bishops
        bishopMoves = moveArray[
            ASEAN_WBISHOP if board.color == WHITE else ASEAN_BBISHOP
        ]
        for cord in iterBits(bishops):
            for c in iterBits(bishopMoves[cord] & notfriends):
                yield newMove(cord, c)

    else:
        # Rooks and Queens and Elephants
        for cord in iterBits(rooks | queens | elephants):
            try:
                attackBoard = (
                    attack00[cord][ray00[cord] & blocker]
                    | attack90[cord][ray90[cord] & blocker]
                )
            except KeyError:
                attackBoard = 0
            for c in iterBits(attackBoard & notfriends):
                yield newMove(cord, c)

        if board.variant == SCHESS and (holding[HAWK] > 0 or holding[ELEPHANT] > 0):
            for cord in iterBits((rooks | queens) & board.virgin[board.color]):
                try:
                    attackBoard = (
                        attack00[cord][ray00[cord] & blocker]
                        | attack90[cord][ray90[cord] & blocker]
                    )
                except KeyError:
                    attackBoard = 0
                for c in iterBits(attackBoard & notfriends):
                    if holding[HAWK] > 0:
                        yield newMove(cord, c, HAWK_GATE)
                    if holding[ELEPHANT] > 0:
                        yield newMove(cord, c, ELEPHANT_GATE)

        # Bishops and Queens, Hawks
        for cord in iterBits(bishops | queens | hawks):
            try:
                attackBoard = (
                    attack45[cord][ray45[cord] & blocker]
                    | attack135[cord][ray135[cord] & blocker]
                )
            except KeyError:
                attackBoard = 0
            for c in iterBits(attackBoard & notfriends):
                yield newMove(cord, c)

        if board.variant == SCHESS and (holding[HAWK] > 0 or holding[ELEPHANT] > 0):
            for cord in iterBits((bishops | queens) & board.virgin[board.color]):
                try:
                    attackBoard = (
                        attack45[cord][ray45[cord] & blocker]
                        | attack135[cord][ray135[cord] & blocker]
                    )
                except KeyError:
                    attackBoard = 0
                for c in iterBits(attackBoard & notfriends):
                    if holding[HAWK] > 0:
                        yield newMove(cord, c, HAWK_GATE)
                    if holding[ELEPHANT] > 0:
                        yield newMove(cord, c, ELEPHANT_GATE)

    # White pawns
    pawnEnemies = enemies | (enpassant is not None and bitPosArray[enpassant] or 0)
    if board.color == WHITE:
        # One step

        if board.variant == SITTUYINCHESS:
            promotion_zone = []
        else:
            promotion_zone = variants[board.variant].PROMOTION_ZONE[WHITE]
        movedpawns = (pawns >> 8) & notblocker  # Move all pawns one step forward
        for cord in iterBits(movedpawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord - 8, cord, p)
            else:
                yield newMove(cord - 8, cord)

        # Two steps

        seccondrow = pawns & rankBits[1]  # Get seccond row pawns
        movedpawns = (seccondrow >> 8) & notblocker  # Move two steps forward, while
        movedpawns = (movedpawns >> 8) & notblocker  # ensuring middle cord is clear
        for cord in iterBits(movedpawns):
            yield newMove(cord - 16, cord)

        # In horde white pawns on first rank may move two squares also
        if board.variant == HORDECHESS:
            firstrow = pawns & rankBits[0]  # Get first row pawns
            movedpawns = (firstrow >> 8) & notblocker  # Move two steps forward, while
            movedpawns = (movedpawns >> 8) & notblocker  # ensuring middle cord is clear
            for cord in iterBits(movedpawns):
                yield newMove(cord - 16, cord)

        # Capture left

        capLeftPawns = pawns & ~fileBits[0]
        capLeftPawns = (capLeftPawns >> 7) & pawnEnemies
        for cord in iterBits(capLeftPawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord - 7, cord, p)
            elif cord == enpassant:
                yield newMove(cord - 7, cord, ENPASSANT)
            else:
                yield newMove(cord - 7, cord)

        # Capture right

        capRightPawns = pawns & ~fileBits[7]
        capRightPawns = (capRightPawns >> 9) & pawnEnemies
        for cord in iterBits(capRightPawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord - 9, cord, p)
            elif cord == enpassant:
                yield newMove(cord - 9, cord, ENPASSANT)
            else:
                yield newMove(cord - 9, cord)

    # Black pawns
    else:
        # One step

        if board.variant == SITTUYINCHESS:
            promotion_zone = []
        else:
            promotion_zone = variants[board.variant].PROMOTION_ZONE[BLACK]
        movedpawns = (pawns << 8) & notblocker
        movedpawns &= 0xFFFFFFFFFFFFFFFF  # contrain to 64 bits
        for cord in iterBits(movedpawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord + 8, cord, p)
            else:
                yield newMove(cord + 8, cord)

        # Two steps

        seccondrow = pawns & rankBits[6]  # Get seventh row pawns
        # Move two steps forward, while ensuring middle cord is clear
        movedpawns = seccondrow << 8 & notblocker
        movedpawns = movedpawns << 8 & notblocker
        for cord in iterBits(movedpawns):
            yield newMove(cord + 16, cord)

        # Capture left

        capLeftPawns = pawns & ~fileBits[7]
        capLeftPawns = capLeftPawns << 7 & pawnEnemies
        for cord in iterBits(capLeftPawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord + 7, cord, p)
            elif cord == enpassant:
                yield newMove(cord + 7, cord, ENPASSANT)
            else:
                yield newMove(cord + 7, cord)

        # Capture right

        capRightPawns = pawns & ~fileBits[0]
        capRightPawns = capRightPawns << 9 & pawnEnemies
        for cord in iterBits(capRightPawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord + 9, cord, p)
            elif cord == enpassant:
                yield newMove(cord + 9, cord, ENPASSANT)
            else:
                yield newMove(cord + 9, cord)

    # Sittuyin promotions
    if board.variant == SITTUYINCHESS and pawns and not queens:
        yield from gen_sittuyin_promotions(board)

    # Cambodian extra first moves for king and queen
    if board.variant == CAMBODIANCHESS:
        if (
            board.arBoard[board.ini_kings[board.color]] == KING
            and board.is_first_move[KING][board.color]
        ):
            if board.color == WHITE:
                if not board.arBoard[B2]:
                    yield newMove(D1, B2)
                if not board.arBoard[F2]:
                    yield newMove(D1, F2)
            else:
                if not board.arBoard[C7]:
                    yield newMove(E8, C7)
                if not board.arBoard[G7]:
                    yield newMove(E8, G7)
        if (
            board.arBoard[board.ini_queens[board.color]] == QUEEN
            and board.is_first_move[QUEEN][board.color]
        ):
            if board.color == WHITE:
                if not board.arBoard[E3]:
                    yield newMove(E1, E3)
            else:
                if not board.arBoard[D6]:
                    yield newMove(D8, D6)

    # Castling
    if kings:
        yield from genCastles(board)


################################################################################
#   Generate capturing moves                                                   #
################################################################################


def genCaptures(board):
    from pychess.Variants import variants

    blocker = board.blocker
    enpassant = board.enpassant

    enemies = board.friends[1 - board.color]

    pawns = board.boards[board.color][PAWN]
    knights = board.boards[board.color][KNIGHT]
    bishops = board.boards[board.color][BISHOP]
    rooks = board.boards[board.color][ROOK]
    queens = board.boards[board.color][QUEEN]
    kings = board.boards[board.color][KING]
    hawks = board.boards[board.color][HAWK]
    elephants = board.boards[board.color][ELEPHANT]

    PROMOTIONS = variants[board.variant].PROMOTIONS
    # In sittuyin promotion can't give capture
    if board.variant == SITTUYINCHESS:
        PROMOTIONS = (NORMAL_MOVE,)
    elif board.variant == LIGHTBRIGADECHESS:
        PROMOTIONS = (QUEEN_PROMOTION,) if board.color == WHITE else (KNIGHT_PROMOTION,)

    # Knights
    knightMoves = moveArray[KNIGHT]
    for cord in iterBits(knights | hawks | elephants):
        for c in iterBits(knightMoves[cord] & enemies):
            yield newMove(cord, c)

    # King
    if kings:
        kingMoves = moveArray[KING]
        # cord = firstBit(kings)
        for cord in iterBits(kings):
            for c in iterBits(kingMoves[cord] & enemies):
                if board.variant != ATOMICCHESS:
                    yield newMove(cord, c)

    # Rooks and Queens
    if board.variant in ASEAN_VARIANTS:
        for cord in iterBits(rooks):
            try:
                attackBoard = (
                    attack00[cord][ray00[cord] & blocker]
                    | attack90[cord][ray90[cord] & blocker]
                )
            except KeyError:
                attackBoard = 0
            for c in iterBits(attackBoard & enemies):
                yield newMove(cord, c)
    else:
        for cord in iterBits(rooks | queens | elephants):
            try:
                attackBoard = (
                    attack00[cord][ray00[cord] & blocker]
                    | attack90[cord][ray90[cord] & blocker]
                )
            except KeyError:
                attackBoard = 0
            for c in iterBits(attackBoard & enemies):
                yield newMove(cord, c)

    # Bishops and Queens
    if board.variant in ASEAN_VARIANTS:
        bishopMoves = moveArray[
            ASEAN_WBISHOP if board.color == WHITE else ASEAN_BBISHOP
        ]
        for cord in iterBits(bishops):
            for c in iterBits(bishopMoves[cord] & enemies):
                yield newMove(cord, c)
        queenMoves = moveArray[ASEAN_QUEEN]
        for cord in iterBits(queens):
            for c in iterBits(queenMoves[cord] & enemies):
                yield newMove(cord, c)
    else:
        for cord in iterBits(bishops | queens | hawks):
            try:
                attackBoard = (
                    attack45[cord][ray45[cord] & blocker]
                    | attack135[cord][ray135[cord] & blocker]
                )
            except KeyError:
                attackBoard = 0
            for c in iterBits(attackBoard & enemies):
                yield newMove(cord, c)

    # White pawns
    pawnEnemies = enemies | (enpassant is not None and bitPosArray[enpassant] or 0)

    if board.color == WHITE:
        promotion_zone = variants[board.variant].PROMOTION_ZONE[WHITE]

        # Promotes

        # Capture left

        capLeftPawns = pawns & ~fileBits[0]
        capLeftPawns = (capLeftPawns >> 7) & pawnEnemies
        for cord in iterBits(capLeftPawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord - 7, cord, p)
            elif cord == enpassant:
                yield newMove(cord - 7, cord, ENPASSANT)
            else:
                yield newMove(cord - 7, cord)

        # Capture right

        capRightPawns = pawns & ~fileBits[7]
        capRightPawns = (capRightPawns >> 9) & pawnEnemies
        for cord in iterBits(capRightPawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord - 9, cord, p)
            elif cord == enpassant:
                yield newMove(cord - 9, cord, ENPASSANT)
            else:
                yield newMove(cord - 9, cord)

    # Black pawns
    else:
        promotion_zone = variants[board.variant].PROMOTION_ZONE[BLACK]

        # One step

        # Capture left

        capLeftPawns = pawns & ~fileBits[7]
        capLeftPawns = capLeftPawns << 7 & pawnEnemies
        for cord in iterBits(capLeftPawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord + 7, cord, p)
            elif cord == enpassant:
                yield newMove(cord + 7, cord, ENPASSANT)
            else:
                yield newMove(cord + 7, cord)

        # Capture right

        capRightPawns = pawns & ~fileBits[0]
        capRightPawns = capRightPawns << 9 & pawnEnemies
        for cord in iterBits(capRightPawns):
            if cord in promotion_zone:
                for p in PROMOTIONS:
                    yield newMove(cord + 9, cord, p)
            elif cord == enpassant:
                yield newMove(cord + 9, cord, ENPASSANT)
            else:
                yield newMove(cord + 9, cord)


################################################################################
#   Generate escapes from check                                                #
################################################################################


def genCheckEvasions(board):
    from pychess.Variants import variants

    color = board.color
    opcolor = 1 - color

    kcord = board.kings[color]
    kings = board.boards[color][KING]
    pawns = board.boards[color][PAWN]
    queens = board.boards[board.color][QUEEN]
    checkers = getAttacks(board, kcord, opcolor)

    arBoard = board.arBoard
    if bin(checkers).count("1") == 1:
        PROMOTIONS = variants[board.variant].PROMOTIONS
        # In sittuyin promotion move not allowed to capture opponent pieces
        if board.variant == SITTUYINCHESS and board.boards[board.color][QUEEN]:
            PROMOTIONS = (NORMAL_MOVE,)
        elif board.variant == LIGHTBRIGADECHESS:
            PROMOTIONS = (
                (QUEEN_PROMOTION,) if board.color == WHITE else (KNIGHT_PROMOTION,)
            )
        promotion_zone = variants[board.variant].PROMOTION_ZONE[color]

        # Captures of checking pieces (except by king, which we will test later)
        chkcord = firstBit(checkers)
        b = getAttacks(board, chkcord, color) & ~kings
        for cord in iterBits(b):
            if not pinnedOnKing(board, cord, color):
                if (
                    arBoard[cord] == PAWN
                    and chkcord in promotion_zone
                    and board.variant != SITTUYINCHESS
                ):
                    for p in PROMOTIONS:
                        yield newMove(cord, chkcord, p)
                else:
                    yield newMove(cord, chkcord)

        # Maybe enpassant can help
        if board.enpassant:
            ep = board.enpassant
            if ep + (color == WHITE and -8 or 8) == chkcord:
                bits = moveArray[color == WHITE and BPAWN or PAWN][ep] & pawns
                for cord in iterBits(bits):
                    if not pinnedOnKing(board, cord, color):
                        yield newMove(cord, ep, ENPASSANT)

        # Lets block/capture the checking piece
        if sliders[arBoard[chkcord]]:
            bits = clearBit(fromToRay[kcord][chkcord], chkcord)

            for cord in iterBits(bits):
                b = getAttacks(board, cord, color)
                b &= ~(kings | pawns)

                # Add in pawn advances
                if color == WHITE and cord > H2:
                    if bitPosArray[cord - 8] & pawns:
                        b |= bitPosArray[cord - 8]
                    if (
                        cord >> 3 == 3
                        and arBoard[cord - 8] == EMPTY
                        and bitPosArray[cord - 16] & pawns
                    ):
                        b |= bitPosArray[cord - 16]

                elif color == BLACK and cord < H7:
                    if bitPosArray[cord + 8] & pawns:
                        b |= bitPosArray[cord + 8]
                    if (
                        cord >> 3 == 4
                        and arBoard[cord + 8] == EMPTY
                        and bitPosArray[cord + 16] & pawns
                    ):
                        b |= bitPosArray[cord + 16]

                for fcord in iterBits(b):
                    # If the piece is blocking another attack, we cannot move it
                    if pinnedOnKing(board, fcord, color):
                        continue
                    if arBoard[fcord] == PAWN and cord in promotion_zone:
                        for p in PROMOTIONS:
                            yield newMove(fcord, cord, p)
                    else:
                        yield newMove(fcord, cord)

                if board.variant == CRAZYHOUSECHESS:
                    holding = board.holding[color]
                    for piece in holding:
                        if holding[piece] > 0:
                            if piece == PAWN:
                                if cord >= 56 or cord <= 7:
                                    continue
                            yield newMove(piece, cord, DROP)

                if board.variant == SITTUYINCHESS and pawns and not queens:
                    from .lmove import TCORD

                    for move in gen_sittuyin_promotions(board):
                        if TCORD(move) == cord:
                            yield move

    # If more than one checkers, move king to get out of check
    if checkers:
        escapes = moveArray[KING][kcord] & ~board.friends[color]
    else:
        escapes = 0

    for chkcord in iterBits(checkers):
        dir = directions[chkcord][kcord]
        if sliders[arBoard[chkcord]]:
            escapes &= ~rays[chkcord][dir]

    for cord in iterBits(escapes):
        if not isAttacked(board, cord, opcolor):
            yield newMove(kcord, cord)


def genDrops(board):
    color = board.color
    arBoard = board.arBoard
    holding = board.holding[color]
    for piece in holding:
        if holding[piece] > 0:
            for cord, elem in enumerate(arBoard):
                if elem == EMPTY:
                    # forbidden drop moves
                    if board.variant == SITTUYINCHESS:
                        if color == WHITE:
                            if cord in (A3, B3, C3, D3) or cord > H3:
                                continue
                            if piece == ROOK and cord > H1:
                                continue
                        else:
                            if cord in (E6, F6, G6, H6) or cord < A6:
                                continue
                            if piece == ROOK and cord < A8:
                                continue

                    elif board.variant == PLACEMENTCHESS:
                        # drop pieces enabled on base line only
                        if color == WHITE:
                            if cord > H1:
                                continue
                        else:
                            if cord < A8:
                                continue

                        # bishops must be on opposite colour squares
                        base_line = arBoard[0:8] if color == WHITE else arBoard[56:64]
                        occupied_colors = [0, 0]
                        occupied_colors[cord % 2] += 1
                        for i, baseline_piece in enumerate(base_line):
                            if baseline_piece != EMPTY:
                                occupied_colors[i % 2] += 1

                        if holding[BISHOP] == 2 and piece != BISHOP:
                            # occupying all same colored fields before any bishop dropped is no-no
                            if (
                                occupied_colors[WHITE] == 4
                                or occupied_colors[BLACK] == 4
                            ):
                                continue
                        elif holding[BISHOP] == 1:
                            for i, baseline_piece in enumerate(base_line):
                                if baseline_piece == BISHOP:
                                    first_bishop_cord = i
                                    break
                            # occupying all possible place of opp colored bishop is no-no
                            if (
                                piece != BISHOP
                                and occupied_colors[1 - first_bishop_cord % 2] == 4
                            ):
                                continue
                            # same colored bishop is no-no
                            elif piece == BISHOP and first_bishop_cord % 2 == cord % 2:
                                continue

                    if piece == PAWN:
                        if cord >= 56 or cord <= 7:
                            continue

                    yield newMove(piece, cord, DROP)



import asyncio
import weakref
from math import ceil

from pychess.Utils.lutils.ldata import MATE_VALUE, MATE_DEPTH


def formatTime(seconds, clk2pgn=False):
    minus = ""
    if seconds <= -10 or seconds >= 10:
        seconds = ceil(seconds)
    if seconds < 0:
        minus = "-"
        seconds = -seconds
    hours, remainder = divmod(seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    if hours or clk2pgn:
        return minus + "%d:%02d:%02d" % (hours, minutes, seconds)
    elif not minutes and seconds < 10:
        return minus + "%.1f" % seconds
    else:
        return minus + "%d:%02d" % (minutes, seconds)


def prettyPrintScore(s, depth, format_mate=False):
    """The score parameter is an eval value from White point of view"""

    # Particular values
    if s is None:
        return "?"
    if s == -MATE_VALUE:
        return _("Illegal")
    if s == 0:
        return "0.00/%s" % depth

    # Preparation
    if s > 0:
        pp = "+"
        mp = ""
    else:
        pp = "-"
        mp = "-"
        s = -s
    if depth:
        depth = "/" + depth
    else:
        depth = ""

    # Rendering
    if s < MATE_VALUE - MATE_DEPTH:
        return f"{pp}{s / 100.0:0.2f}{depth}"
    else:
        mate_in = int(MATE_VALUE - s)
        if format_mate:
            if mate_in == 0:
                return _("Mate")
            return "%s #%s%d" % (_("Mate"), mp, mate_in)
        else:
            return f"{pp}#{s:.0f}"  # Sign before sharp to be parsed in PGN


def createStoryTextAppEvent(text):
    try:
        import storytext

        storytext.applicationEvent(text)
    except AttributeError:
        pass
    except ImportError:
        pass


class wait_signal(asyncio.Future):
    """A future for waiting for a given signal to occur."""

    def __init__(self, obj, name, *, loop=None):
        super().__init__(loop=loop)
        self._obj = weakref.ref(obj, lambda s: self.cancel())
        self._hnd = obj.connect(name, self._signal_callback)

    def _signal_callback(self, *k):
        obj = self._obj()
        if obj is not None:
            obj.disconnect(self._hnd)
        self.set_result(k)

    def cancel(self, msg=None):
        if self.cancelled():
            return False
        try:
            super().cancel(msg=msg)
        except TypeError:  # It has msg parameter only form Python 3.9
            super().cancel()
        except AttributeError:
            pass
        try:
            obj = self._obj()
            if obj is not None:
                obj.disconnect(self._hnd)
        except AttributeError:
            pass
        return True



import os
import shutil
import time

from sqlalchemy import (
    create_engine,
    MetaData,
    Table,
    Column,
    Integer,
    String,
    SmallInteger,
    ForeignKey,
    event,
    select,
)
from sqlalchemy.engine import Engine
from sqlalchemy.pool import StaticPool
from sqlalchemy.exc import OperationalError

# from sqlalchemy.ext.compiler import compiles
# from sqlalchemy.sql.expression import Executable, ClauseElement, _literal_as_text

from pychess.System.Log import log
from pychess.System.prefix import addUserCachePrefix


@event.listens_for(Engine, "before_cursor_execute")
def before_cursor_execute(conn, cursor, statement, parameters, context, executemany):
    context._query_start_time = time.time()
    log.debug("Start Query:\n%s" % statement, extra={"task": "SQL"})
    log.debug(f"Parameters:\n{parameters!r}", extra={"task": "SQL"})


@event.listens_for(Engine, "after_cursor_execute")
def after_cursor_execute(conn, cursor, statement, parameters, context, executemany):
    total = time.time() - context._query_start_time
    log.debug("Query Complete!", extra={"task": "SQL"})
    log.debug("Total Time: %.02fms" % (total * 1000), extra={"task": "SQL"})


# Just to make sphinx happy...
# try:
#     class Explain(Executable, ClauseElement):
#         def __init__(self, stmt):
#             self.statement = _literal_as_text(stmt)
# except TypeError:
#     class Explain:
#         pass


# @compiles(Explain, 'sqlite')
# def slite_explain(element, compiler, **kw):
#     text = "EXPLAIN QUERY PLAN "
#     text += compiler.process(element.statement, **kw)
#     return text


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    # www.sqlite.org/pragma.html
    cursor.execute("PRAGMA page_size = 4096")
    cursor.execute("PRAGMA cache_size=10000")
    # cursor.execute("PRAGMA locking_mode=EXCLUSIVE")
    cursor.execute("PRAGMA synchronous=NORMAL")
    # cursor.execute("PRAGMA journal_mode=WAL")
    # cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()


def insert_or_ignore(engine, stmt):
    if engine.name == "sqlite":
        # can't use "OR UPDATE" because it delete+insert records
        # and breaks referential integrity
        return stmt.prefix_with("OR IGNORE")
    elif engine.name == "postgresql":
        return stmt.prefix_with("ON CONFLICT DO NOTHING")
    elif engine.name == "mysql":
        return stmt.prefix_with("IGNORE")


engines = {}

# PyChess database schema version
SCHEMA_VERSION = "20180221"


def get_schema_version(engine):
    with engine.connect() as connection:
        result = connection.execute(select(schema_version.c.version)).scalar()
    return result


def get_engine(path=None, dialect="sqlite", echo=False):
    if path is None:
        # In memory database
        url = "sqlite://"
    elif dialect == "sqlite":
        url = f"{dialect}:///{path}"

    if url in engines and os.path.isfile(path) and os.path.getsize(path) > 0:
        return engines[url]
    else:
        if path is None:
            engine = create_engine(
                url,
                connect_args={"check_same_thread": False},
                echo=echo,
                poolclass=StaticPool,
            )
        else:
            if path != empty_db and (
                not os.path.isfile(path) or os.path.getsize(path) == 0
            ):
                shutil.copyfile(empty_db, path)
            engine = create_engine(url, echo=echo)

        if path != empty_db and (
            path is None or get_schema_version(engine) != SCHEMA_VERSION
        ):
            metadata.drop_all(engine)
            metadata.create_all(engine)
            ini_schema_version(engine)

        engines[url] = engine
        return engine


metadata = MetaData()

source = Table(
    "source",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("name", String(256)),
    Column("info", String(256)),
)

event = Table(
    "event",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("name", String(256), index=True),
)

site = Table(
    "site",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("name", String(256), index=True),
)

annotator = Table(
    "annotator",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("name", String(256), index=True),
)

player = Table(
    "player",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("name", String(256), index=True),
)

pl1 = player.alias()
pl2 = player.alias()

game = Table(
    "game",
    metadata,
    Column("id", Integer, primary_key=True, index=True),
    Column("offset", Integer, index=True),
    Column("offset8", Integer, index=True),
    Column("event_id", Integer, ForeignKey("event.id"), index=True),
    Column("site_id", Integer, ForeignKey("site.id"), index=True),
    Column("date", String(10), default=""),
    Column("round", String(8), default=""),
    Column("white_id", Integer, ForeignKey("player.id"), index=True),
    Column("black_id", Integer, ForeignKey("player.id"), index=True),
    Column("result", SmallInteger, default=0),
    Column("white_elo", String(4), default=""),
    Column("black_elo", String(4), default=""),
    Column("ply_count", String(3), default=""),
    Column("eco", String(3), default=""),
    Column("time_control", String(7), default=""),
    Column("board", SmallInteger, default=0),
    Column("fen", String(128), default=""),
    Column("variant", SmallInteger, default=0),
    Column("annotator_id", Integer, ForeignKey("annotator.id"), index=True),
    Column("source_id", Integer, ForeignKey("source.id"), index=True),
)

tag_game = Table(
    "tag_game",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("game_id", Integer, ForeignKey("game.id"), nullable=False),
    Column("tag_name", String(128), default=""),
    Column("tag_value", String(128), default=""),
)

schema_version = Table(
    "schema_version",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("version", String(8)),
)


def drop_indexes(engine):
    for table in metadata.tables.values():
        for index in table.indexes:
            try:
                index.drop(bind=engine)
            except OperationalError as e:
                if e.orig.args[0].startswith("no such index"):
                    pass
                    # print(e.orig.args[0])
                else:
                    raise


def create_indexes(engine):
    for table in metadata.tables.values():
        for index in table.indexes:
            index.create(bind=engine)


def ini_schema_version(engine):
    conn = engine.connect()
    conn.execute(
        schema_version.insert(),
        [
            {"id": 1, "version": SCHEMA_VERSION},
        ],
    )
    conn.close()


# create an empty database to use as skeleton
empty_db = os.path.join(addUserCachePrefix("%s.sqlite" % SCHEMA_VERSION))
if not os.path.isfile(empty_db):
    engine = get_engine(empty_db)
    metadata.create_all(engine)
    ini_schema_version(engine)
    engine.dispose()



from math import ceil as float_ceil, pi
import cairo

from gi.repository import Gtk, Gdk, GObject

from .OverlayWindow import OverlayWindow

# ceil = lambda f: int(float_ceil(f))


def ceil(num):
    return int(float_ceil(num))


POSITIONS_COUNT = 5
NORTH, EAST, SOUTH, WEST, CENTER = range(POSITIONS_COUNT)
DX_DY = ((0, -1), (1, 0), (0, 1), (-1, 0), (0, 0))
PADDING_X = 0.2  # Amount of button width
PADDING_Y = 0.4  # Amount of button height


class StarArrowButton(OverlayWindow):
    __gsignals__ = {
        "dropped": (GObject.SignalFlags.RUN_FIRST, None, (int, object)),
        "hovered": (GObject.SignalFlags.RUN_FIRST, None, (int, object)),
        "left": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self, parent, northSvg, eastSvg, southSvg, westSvg, centerSvg, bgSvg):
        OverlayWindow.__init__(self, parent)

        self.svgs = (northSvg, eastSvg, southSvg, westSvg, centerSvg)
        self.bgSvg = bgSvg
        self.size = ()
        self.currentHovered = -1

        # targets = [("GTK_NOTEBOOK_TAB", Gtk.TargetFlags.SAME_APP, 0xbadbeef)]
        targets = [
            Gtk.TargetEntry.new("GTK_NOTEBOOK_TAB", Gtk.TargetFlags.SAME_APP, 0xBADBEEF)
        ]
        self.drag_dest_set(
            Gtk.DestDefaults.DROP | Gtk.DestDefaults.MOTION,
            targets,
            Gdk.DragAction.MOVE,
        )
        self.drag_dest_set_track_motion(True)
        self.myparent.button_cids += [
            self.connect("drag-motion", self.__onDragMotion),
            self.connect("drag-leave", self.__onDragLeave),
            self.connect("drag-drop", self.__onDragDrop),
            self.connect_after("draw", self.__onExposeEvent),
        ]
        self.myparentAlloc = None
        self.myparentPos = None
        self.hasHole = False
        self.size = ()

    def _calcSize(self):
        parentAlloc = self.myparent.get_allocation()

        if (
            self.myparentAlloc is None
            or parentAlloc.width != self.myparentAlloc.width
            or parentAlloc.height != self.myparentAlloc.height
        ):
            starWidth, starHeight = self.getSizeOfSvg(self.bgSvg)
            scale = min(
                1,
                parentAlloc.width / float(starWidth),
                parentAlloc.height / float(starHeight),
            )
            self.size = list(map(int, (starWidth * scale, starHeight * scale)))
            self.resize(self.size[0], self.size[1])

            if self.get_window():
                self.hasHole = True
                self.digAHole(self.bgSvg, self.size[0], self.size[1])

        elif not self.hasHole:
            self.hasHole = True
            self.digAHole(self.bgSvg, self.size[0], self.size[1])

        if self.myparent.get_window():
            x_loc, y_loc = self.translateCoords(
                int(parentAlloc.width / 2.0 - self.size[0] / 2.0),
                int(parentAlloc.height / 2.0 - self.size[1] / 2.0),
            )
            if (x_loc, y_loc) != self.get_position():
                self.move(x_loc, y_loc)

            self.myparentPos = self.myparent.get_window().get_position()
        self.myparentAlloc = parentAlloc

    def __onExposeEvent(self, self_, ctx):
        self._calcSize()

        context = self.get_window().cairo_create()
        self.paintTransparent(context)
        surface = self.getSurfaceFromSvg(self.bgSvg, self.size[0], self.size[1])
        context.set_source_surface(surface, 0, 0)
        context.paint()

        for position in range(POSITIONS_COUNT):
            rect = self.__getButtonRectangle(position)

            context = self.get_window().cairo_create()
            surface = self.getSurfaceFromSvg(
                self.svgs[position], rect.width, rect.height
            )
            context.set_source_surface(surface, rect.x, rect.y)
            context.paint()

    def __getButtonRectangle(self, position):
        starWidth, starHeight = self.getSizeOfSvg(self.bgSvg)
        buttonWidth, buttonHeight = self.getSizeOfSvg(self.svgs[position])

        buttonWidth = buttonWidth * self.size[0] / float(starWidth)
        buttonHeight = buttonHeight * self.size[1] / float(starHeight)
        dx_loc, dy_loc = DX_DY[position]
        x_loc = ceil(
            dx_loc * (1 + PADDING_X) * buttonWidth
            - buttonWidth / 2.0
            + self.size[0] / 2.0
        )
        y_loc = ceil(
            dy_loc * (1 + PADDING_Y) * buttonHeight
            - buttonHeight / 2.0
            + self.size[1] / 2.0
        )

        rect = Gdk.Rectangle()
        rect.x, rect.y, rect.width, rect.height = (
            x_loc,
            y_loc,
            ceil(buttonWidth),
            ceil(buttonHeight),
        )
        return rect

    def __getButtonAtPoint(self, x, y):
        for position in range(POSITIONS_COUNT):
            rect = Gdk.Rectangle()
            rect.x, rect.y, rect.width, rect.height = (x, y, 1, 1)

            inside, dest = Gdk.rectangle_intersect(
                self.__getButtonRectangle(position), rect
            )
            if inside:
                return position
        return -1

    def __onDragMotion(self, arrow, context, x, y, timestamp):
        position = self.__getButtonAtPoint(x, y)
        if self.currentHovered != position:
            self.currentHovered = position
            if position > -1:
                self.emit("hovered", position, Gtk.drag_get_source_widget(context))
            else:
                self.emit("left")

        if position > -1:
            Gdk.drag_status(context, Gdk.DragAction.MOVE, timestamp)
            return True
        Gdk.drag_status(context, Gdk.DragAction.DEFAULT, timestamp)

    def __onDragLeave(self, arrow, context, timestamp):
        if self.currentHovered != -1:
            self.currentHovered = -1
            self.emit("left")

    def __onDragDrop(self, arrow, context, x, y, timestamp):
        position = self.__getButtonAtPoint(x, y)
        if position > -1:
            self.emit("dropped", position, Gtk.drag_get_source_widget(context))
            context.finish(True, True, timestamp)
            return True


if __name__ == "__main__":
    w = Gtk.Window()
    w.connect("delete-event", Gtk.main_quit)
    sab = StarArrowButton(
        w,
        "/home/thomas/Programmering/workspace/pychess/glade/dock_top.svg",
        "/home/thomas/Programmering/workspace/pychess/glade/dock_right.svg",
        "/home/thomas/Programmering/workspace/pychess/glade/dock_bottom.svg",
        "/home/thomas/Programmering/workspace/pychess/glade/dock_left.svg",
        "/home/thomas/Programmering/workspace/pychess/glade/dock_center.svg",
        "/home/thomas/Programmering/workspace/pychess/glade/dock_star.svg",
    )

    def on_expose(widget, event):
        cairo_win = widget.window.cairo_create()
        cx_loc = cy_loc = 100
        radius = 50
        cairo_win.arc(cx_loc, cy_loc, radius - 1, 0, 2 * pi)
        cairo_win.set_source_rgba(1.0, 0.0, 0.0, 1.0)
        cairo_win.set_operator(cairo.OPERATOR_OVER)
        cairo_win.fill()

    # w.connect("e)
    w.show_all()
    sab.show_all()
    Gtk.main()

# if __name__ != "__main__":
#    w = Gtk.Window()
#    w.connect("delete-event", Gtk.main_quit)
#    hbox = Gtk.HBox()
#
#    l = Gtk.Layout()
#    l.set_size_request(200,200)
#    sab = StarArrowButton("/home/thomas/Programmering/workspace/pychess/glade/dock_top.svg",
#                          "/home/thomas/Programmering/workspace/pychess/glade/dock_right.svg",
#                          "/home/thomas/Programmering/workspace/pychess/glade/dock_bottom.svg",
#                          "/home/thomas/Programmering/workspace/pychess/glade/dock_left.svg",
#                          "/home/thomas/Programmering/workspace/pychess/glade/dock_center.svg",
#                          "/home/thomas/Programmering/workspace/pychess/glade/dock_star.svg")
#    sab.set_size_request(200,200)
#    l.put(sab, 0, 0)
#    hbox.add(l)
#    def handle (*args):
#        sab.showAt(l, CENTER)
#    l.connect("button-press-event", handle)
#
#    nb = Gtk.Notebook()
#    label = Gtk.Label(label="hi")
#    nb.append_page(label)
#    nb.set_tab_detachable(label, True)
#    hbox.add(nb)
#    w.add(hbox)
#    w.show_all()
#    Gtk.main()



import asyncio
import os

from pychess.System.prefix import addDataPrefix
from pychess.Utils.const import (
    WHITE,
    BLACK,
    LOCAL,
    NORMALCHESS,
    ARTIFICIAL,
    WAITING_TO_START,
    HINT,
    PRACTICE_GOAL_REACHED,
    PUZZLE,
)
from pychess.Utils.LearnModel import LearnModel
from pychess.Utils.TimeModel import TimeModel
from pychess.Variants import variants
from pychess.Players.Human import Human
from pychess.Players.engineNest import discoverer
from pychess.perspectives import perspective_manager
from pychess.perspectives.learn.generateLessonsSidepanel import generateLessonsSidepanel
from pychess.perspectives.learn import lessons_solving_progress
from pychess.perspectives.learn import puzzles_solving_progress
from pychess.Savers.olv import OLVFile
from pychess.Savers.pgn import PGNFile
from pychess.System import conf
from pychess.System.protoopen import protoopen

__title__ = _("Puzzles")

__icon__ = addDataPrefix("glade/panel_book.svg")

__desc__ = _("Lichess practice studies Puzzles from GM games and Chess compositions")


# https://lichess.org/practice, http://wtharvey.com, http://www.yacpdb.org

puzzles0 = []
puzzles1 = []
puzzles2 = []
puzzles3 = []
for elem in sorted(os.listdir(path=addDataPrefix("learn/puzzles/"))):
    if elem.startswith("lichess_study") and elem.endswith(".pgn"):
        if elem[14:31] == "lichess-practice-":
            puzzles0.append(
                (
                    elem,
                    elem[31 : elem.find("_by_")].replace("-", " ").capitalize(),
                    "lichess.org",
                )
            )
        else:
            puzzles0.append(
                (
                    elem,
                    elem[14 : elem.find("_by_").replace("-", " ").capitalize()],
                    "lichess.org",
                )
            )
    elif elem.startswith("mate_in_") and elem.endswith(".pgn"):
        puzzles1.append((elem, "Puzzles by GMs: Mate in %s" % elem[8], "wtharvey.com"))
    elif elem.endswith(".olv"):
        puzzles2.append(
            (elem, "Puzzles by %s" % elem.split(".olv")[0].capitalize(), "yacpdb.org")
        )
    elif elem.endswith(".pgn"):
        puzzles3.append((elem, elem.split(".pgn")[0].capitalize(), _("others")))

PUZZLES = puzzles0 + puzzles1 + puzzles2 + puzzles3

# Note: Find the declaration of the class Sidepanel at the end of the file


HINTS = {
    "r2qrbk1/5ppp/pn1p4/np2P1P1/3p4/5N2/PPB2PP1/R1BQR1K1 w - - 1 20": (38, "c2h7"),
    "3r1rk1/bpq2ppp/p1b1p3/2P5/1P2B3/P4Q2/1B3PPP/2R2RK1 w - - 3 18": (34, "e4h7"),
    "8/8/3p1p2/1N2p3/4kNQ1/5n2/2K3B1/8 w - - 0 1": (0, "g2h1"),
}


def start_puzzle_from(filename, index=None):
    if filename.lower().endswith(".pgn"):
        chessfile = PGNFile(protoopen(addDataPrefix("learn/puzzles/%s" % filename)))
        chessfile.limit = 1000
        chessfile.init_tag_database()
    elif filename.lower().endswith(".olv"):
        chessfile = OLVFile(
            protoopen(addDataPrefix("learn/puzzles/%s" % filename), encoding="utf-8")
        )

    records, plys = chessfile.get_records()

    progress = puzzles_solving_progress.get(filename, [0] * chessfile.count)

    if index is None:
        try:
            index = progress.index(0)
        except ValueError:
            index = 0

    rec = records[index]

    timemodel = TimeModel(0, 0)
    gamemodel = LearnModel(timemodel)

    chessfile.loadToModel(rec, 0, gamemodel)

    start_puzzle_game(gamemodel, filename, records, index, rec)


def start_puzzle_game(gamemodel, filename, records, index, rec, from_lesson=False):
    gamemodel.set_learn_data(
        PUZZLE, filename, index, len(records), from_lesson=from_lesson
    )

    if "FEN" in gamemodel.tags and gamemodel.tags["FEN"] in HINTS:
        gamemodel.add_hints(HINTS[gamemodel.tags["FEN"]])

    engine = discoverer.getEngineByName(discoverer.getEngineLearn())
    ponder_off = True

    color = gamemodel.boards[0].color

    w_name = "" if rec["White"] is None else rec["White"]
    b_name = "" if rec["Black"] is None else rec["Black"]

    player_name = conf.get("firstName")
    engine_name = discoverer.getName(engine)

    if rec["Event"].startswith("Lichess Practice"):
        w_name = player_name if color == WHITE else engine_name
        b_name = engine_name if color == WHITE else player_name

    opp_name = engine_name if rec["Event"].startswith("Lichess Practice") else b_name

    if color == WHITE:
        p0 = (LOCAL, Human, (WHITE, w_name), w_name)
        p1 = (
            ARTIFICIAL,
            discoverer.initPlayerEngine,
            (engine, BLACK, 20, variants[NORMALCHESS], 20, 0, 0, ponder_off),
            b_name,
        )
    else:
        p0 = (
            ARTIFICIAL,
            discoverer.initPlayerEngine,
            (engine, WHITE, 20, variants[NORMALCHESS], 20, 0, 0, ponder_off),
            w_name,
        )
        p1 = (LOCAL, Human, (BLACK, b_name), b_name)

    def on_game_started(gamemodel, name, color):
        perspective.activate_panel("annotationPanel")
        asyncio.create_task(
            gamemodel.start_analyzer(HINT, force_engine=discoverer.getEngineLearn())
        )
        gamemodel.players[1 - color].name = name
        gamemodel.emit("players_changed")

    gamemodel.connect("game_started", on_game_started, opp_name, color)

    def goal_checked(gamemodle):
        if gamemodel.reason == PRACTICE_GOAL_REACHED:
            if from_lesson:
                progress = lessons_solving_progress[gamemodel.source]
            else:
                progress = puzzles_solving_progress[gamemodel.source]

            progress[gamemodel.current_index] = 1

            if from_lesson:
                lessons_solving_progress[gamemodel.source] = progress
            else:
                puzzles_solving_progress[gamemodel.source] = progress

    gamemodel.connect("goal_checked", goal_checked)

    gamemodel.variant.need_initial_board = True
    gamemodel.status = WAITING_TO_START

    perspective = perspective_manager.get_perspective("games")
    asyncio.create_task(perspective.generalStart(gamemodel, p0, p1))


# Sidepanel is a class
Sidepanel = generateLessonsSidepanel(
    solving_progress=puzzles_solving_progress,
    learn_category_id=PUZZLE,
    entries=PUZZLES,
    start_from=start_puzzle_from,
)



from gi.repository import GLib, GObject

from pychess.ic import GAME_TYPES
from pychess.ic.icc import (
    DG_PLAYERS_IN_MY_GAME,
    DG_KIBITZ,
    DG_PERSONAL_TELL,
    DG_SHOUT,
    DG_CHANNEL_TELL,
    DG_PEOPLE_IN_MY_CHANNEL,
    DG_CHANNELS_SHARED,
)
from pychess.ic.managers.ChatManager import ChatManager

CHANNEL_SHOUT = "shout"
CHANNEL_CSHOUT = "cshout"


class ICCChatManager(ChatManager):
    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection

        self.connection.expect_dg_line(
            DG_PLAYERS_IN_MY_GAME, self.on_icc_players_in_my_game
        )
        self.connection.expect_dg_line(DG_KIBITZ, self.on_icc_kibitz)
        self.connection.expect_dg_line(DG_PERSONAL_TELL, self.on_icc_personal_tell)
        self.connection.expect_dg_line(DG_SHOUT, self.on_icc_shout)
        self.connection.expect_dg_line(DG_CHANNEL_TELL, self.on_icc_channel_tell)
        self.connection.expect_dg_line(
            DG_PEOPLE_IN_MY_CHANNEL, self.on_icc_people_in_my_channel
        )
        self.connection.expect_dg_line(DG_CHANNELS_SHARED, self.on_icc_channels_shared)

        self.connection.client.run_command("set-2 %s 1" % DG_PLAYERS_IN_MY_GAME)
        self.connection.client.run_command("set-2 %s 1" % DG_KIBITZ)
        self.connection.client.run_command("set-2 %s 1" % DG_PERSONAL_TELL)
        self.connection.client.run_command("set-2 %s 1" % DG_SHOUT)
        self.connection.client.run_command("set-2 %s 1" % DG_CHANNEL_TELL)
        self.connection.client.run_command("set-2 %s 1" % DG_PEOPLE_IN_MY_CHANNEL)
        self.connection.client.run_command("set-2 %s 1" % DG_CHANNELS_SHARED)

        self.currentLogChannel = None

        self.connection.client.run_command("set Lang English")

        self.connection.client.run_command("set height 240")

        self.connection.client.run_command("inchannel %s" % self.connection.username)
        self.connection.client.run_command("help channel_list")
        self.observers = {}

        self.channels = [
            (CHANNEL_SHOUT, _("Shout")),
            (CHANNEL_CSHOUT, _("Chess Shout")),
        ]

        for id, channel in ICC_CHANNELS:
            self.channels.append((str(id), channel))
        GLib.idle_add(self.emit, "channelsListed", self.channels)

    def on_icc_players_in_my_game(self, data):
        # gamenumber playername symbol kibvalue
        # O=observing
        # PW=playing white
        # PB=playing black
        # SW=playing simul and is white
        # SB=playing simul and is black
        # E=Examining
        # X=None (has left the table)

        gameno, name, symbol, kibvalue = data.split()

        ficsplayer = self.connection.players.get(name)
        rating = ficsplayer.getRatingByGameType(GAME_TYPES["standard"])
        if rating:
            name = f"{name}({rating})"

        if gameno not in self.observers:
            observers = set()
            self.observers[gameno] = observers

        if symbol == "O":
            self.observers[gameno].add(name)
        elif symbol == "X" and name in self.observers[gameno]:
            self.observers[gameno].remove(name)

        obs_str = " ".join(list(self.observers[gameno]))
        self.emit("observers_received", gameno, obs_str)

    def on_icc_kibitz(self, data):
        # gamenumber playername titles kib/whi ^Y{kib string^Y}
        gameno, name, rest = data.split(" ", 2)
        titles, rest = rest.split("}", 1)
        kib_whi, text = rest[1:].split(" ", 1)
        text = text[2:-2]

        if kib_whi == "1":
            GLib.idle_add(self.emit, "kibitzMessage", name, int(gameno), text)
        else:
            GLib.idle_add(self.emit, "whisperMessage", name, int(gameno), text)

    def on_icc_personal_tell(self, data):
        # playername titles ^Y{tell string^Y} type
        name, rest = data.split(" ", 1)
        titles, rest = rest.split("}", 1)
        text, tell_type = rest[3:].split("}")
        isadmin = tell_type == "4"
        GLib.idle_add(self.emit, "privateMessage", name, "title", isadmin, text[:-1])

    def on_icc_shout(self, data):
        # playername titles type ^Y{shout string^Y}
        name, rest = data.split(" ", 1)
        titles, rest = rest.split("}", 1)
        shout_type, text = rest[1:].split("{")
        # print(name, shout_type, text)
        isadmin = shout_type == "3"  # announcements
        isme = name.lower() == self.connection.username.lower()
        GLib.idle_add(
            self.emit, "channelMessage", name, isadmin, isme, "shout", text[:-2]
        )

    def on_icc_channel_tell(self, data):
        # channel playername titles ^Y{tell string^Y} type
        channel, name, rest = data.split(" ", 2)
        titles, rest = rest.split("}", 1)
        text, tell_type = rest[3:].split("}")
        isme = name.lower() == self.connection.username.lower()
        isadmin = tell_type == "4"
        GLib.idle_add(self.emit, "channelMessage", name, isadmin, isme, channel, text)

    def on_icc_people_in_my_channel(self, data):
        # channel playername come/go
        print(data)

    def on_icc_channels_shared(self, data):
        # playername channels
        print(data)


ICC_CHANNELS = (
    (0, "Admins only"),
    (1, "Newbie Help"),
    (2, "Experienced Help"),
    (3, "Simul"),
    (4, "SimulBot2"),
    (5, "Wild 5"),
    (7, "Wild 7"),
    (10, "Team-Setup"),
    (11, "Team Game Channel A"),
    (12, "Team Game Channel B"),
    (14, "Macintosh users channel"),
    (22, "Fischer-Random Chess"),
    (23, "Crazyhouse Channel"),
    (24, "BUGHOUSE Channel"),
    (25, "3 checks you win"),
    (26, "Giveaway Chess"),
    (27, "Atomic Channel"),
    (28, "Shatranj Channel"),
    (32, "STtourney channel"),
    (34, "Sports"),
    (43, "Chess theory"),
    (46, "Tomato TD"),
    (47, "Tomato tournament managers"),
    (49, "Flash TD"),
    (64, "Computer chess"),
    (65, "Canadian"),
    (66, "Australia"),
    (67, "British"),
    (68, "South Africa"),
    (69, "Singapore and Malaysia"),
    (70, "Greek"),
    (71, "Spanish"),
    (72, "French"),
    (73, "German"),
    (74, "Dutch"),
    (75, "Russian"),
    (76, "Italian"),
    (77, "Japanese"),
    (78, "Scandinavian(Denmark,Norway,Sweden)"),
    (79, "Icelandic"),
    (80, "Finnish"),
    (81, "Portuguese"),
    (82, "Catalan"),
    (83, "Hebrew"),
    (84, "Turkish"),
    (86, "ASCII art"),
    (88, "China"),
    (90, "Slow time controls"),
    (97, "Politics"),
    (100, "Admins and helpers only"),
    (101, "Music"),
    (103, "Religious Discussions/Debates"),
    (107, "Math and Science"),
    (110, "Philosophy"),
    (114, "Health and Medicine"),
    (116, "Wild 16 (Kriegspiel)"),
    (117, "Wild 17 (Losers Chess)"),
    (123, "Acrobot"),
    (129, "USCL (United States Chess League)"),
    (147, "Tomato senior managers group"),
    (165, "ICC Radio Broadcast and Chess Events"),
    (183, "Correspondence Chess"),
    (185, "Philippine"),
    (186, "Irish"),
    (220, "Query all the Tomato-type TD bots"),
    (221, "The 'Ask for a Tournament' Channel"),
    (222, "Slomato TD (Slow Tournaments)"),
    (223, "WildOne TD (Wild Tournaments)"),
    (224, "Cooly TD (Tournaments)"),
    (225, "LittlePer TD (Tournaments)"),
    (226, "Automato TD (Special Event Tournaments)"),
    (227, "Pear TD (Special Event Tournaments)"),
    (228, "Ketchup TD (Tournaments)"),
    (229, "Channel for Tournament Robots to interact"),
    (230, "Olive TD (Tournaments)"),
    (232, "Yenta TD (Special Event Tournaments)"),
    (250, "Lobby"),
    (271, "LatinTrivia Channel (Spanish)"),
    (272, "TriviaBot Channel"),
    (274, "Spam channel"),
    (280, "BettingBot"),
    (291, "Scholastic Chess Coaches"),
    (300, "Helpers and Administrators"),
    (302, "Busters"),
    (303, "Christian"),
    (333, "AtheistAgnostic"),
    (334, "Johns Hopkins CTY"),
    (335, "SimulBot"),
    (337, "Broadcast"),
    (338, "Relay coordination"),
    (340, "LeChessClub"),
    (345, "Team4545League"),
    (348, "Tomato Trainers group"),
    (394, "Tomato admins group"),
    (396, "Complain"),
    (397, "Politics-Reserved"),
    (399, "Events"),
)



import re
from math import floor

from gi.repository import Gtk
from gi.repository import Pango
from gi.repository import Gdk

from pychess.Utils import prettyPrintScore
from pychess.Utils.const import (
    WHITE,
    BLACK,
    FEN_EMPTY,
    reprResult,
    reprSign,
    FAN_PIECES,
)
from pychess.System import conf
from pychess.System.prefix import addDataPrefix
from pychess.Utils.Cord import Cord
from pychess.Utils.lutils.LBoard import LBoard
from pychess.Utils.lutils.lmove import toSAN, toFAN, FCORD, TCORD
from pychess.Savers.pgn import move_count, nag2symbol, parseTimeControlTag
from pychess.widgets.ChessClock import formatTime
from pychess.widgets.LearnInfoBar import LearnInfoBar
from pychess.widgets import insert_formatted, preferencesDialog, mainwindow
from pychess.widgets.Background import isDarkTheme


# --- Constants

__title__ = _("Annotation")
__active__ = True
__icon__ = addDataPrefix("glade/panel_annotation.svg")
__desc__ = _("Annotated game")

EMPTY_BOARD = LBoard()
EMPTY_BOARD.applyFen(FEN_EMPTY)


css = """
GtkButton#rounded {
    border-radius: 20px;
}
"""


def add_provider(widget):
    screen = widget.get_screen()
    style = widget.get_style_context()
    provider = Gtk.CssProvider()
    provider.load_from_data(css.encode("utf-8"))
    style.add_provider_for_screen(screen, provider, Gtk.STYLE_PROVIDER_PRIORITY_USER)


# -- Documentation
"""
We are maintaining a list of nodes to help manipulate the textbuffer.
Node can represent a move, comment or variation (start/end) marker.
Nodes are dicts with keys like:
board  = in move node it's the lboard of move
         in comment node it's the lboard where the comment belongs to
         in end variation marker node it's the first lboard of the variation
         in start variation marker is's None
start  = the beginning offest of the node in the textbuffer
end    = the ending offest of the node in the textbuffer
parent = the parent lboard if the node is a move in a variation, otherwise None
vari   = in end variation node it's the start variation marker node
         in start variation node it's None
level  = depth in variation tree (0 for mainline nodes, 1 for first level variation moves, etc.)
index  = in comment nodes the index of comment if more exist for a move
"""


# -- Widget


class Sidepanel:
    def load(self, gmwidg):
        """
        The method initializes the widget, attached events, internal variables, layout...
        """

        # Internal variables
        self.nodelist = []
        self.boardview = gmwidg.board.view
        self.gamemodel = gmwidg.gamemodel
        self.variation_to_remove = None
        if self.gamemodel is None:
            return None

        # Internal objects/helpers
        self.cursor_standard = Gdk.Cursor.new(Gdk.CursorType.LEFT_PTR)
        self.cursor_hand = Gdk.Cursor.new(Gdk.CursorType.HAND2)

        # Header text area
        self.header_textview = Gtk.TextView()
        self.header_textview.set_wrap_mode(Gtk.WrapMode.WORD)
        self.header_textview.set_editable(False)
        self.header_textview.set_cursor_visible(False)
        self.header_textbuffer = self.header_textview.get_buffer()

        # Header text tags
        self.header_textbuffer.create_tag("head1")
        self.header_textbuffer.create_tag("head2", weight=Pango.Weight.BOLD)

        # Move text area
        self.textview = Gtk.TextView()
        self.textview.set_wrap_mode(Gtk.WrapMode.WORD)
        self.textbuffer = self.textview.get_buffer()

        # Load of the preferences
        def cb_config_changed(*args):
            self.fetch_chess_conf()
            self.tag_move.set_property("font_desc", self.font)
            for i in range(len(self.tag_vari_depth)):
                self.tag_vari_depth[i].set_property("font_desc", self.font)
            self.update()

        self.fetch_chess_conf()

        self.cids_conf = []
        self.cids_conf.append(conf.notify_add("movetextFont", cb_config_changed))
        self.cids_conf.append(conf.notify_add("figuresInNotation", cb_config_changed))
        self.cids_conf.append(conf.notify_add("showEmt", cb_config_changed))
        self.cids_conf.append(conf.notify_add("showBlunder", cb_config_changed))
        self.cids_conf.append(conf.notify_add("showEval", cb_config_changed))

        # Move text tags
        self.tag_remove_variation = self.textbuffer.create_tag("remove-variation")
        self.tag_new_line = self.textbuffer.create_tag("new_line")

        self.tag_move = self.textbuffer.create_tag("move", font_desc=self.font)
        palette = self.get_palette()
        self.tag_vari_depth = []
        for i in range(64):
            tag = self.textbuffer.create_tag(
                "variation-depth-%d" % i,
                font_desc=self.font,
                foreground=palette[i % len(palette)],
                style="italic",
                left_margin=15 * (i + 1),
            )
            self.tag_vari_depth.append(tag)

        self.textbuffer.create_tag("scored0")
        self.textbuffer.create_tag("scored1", foreground_rgba=Gdk.RGBA(0.2, 0, 0, 1))
        self.textbuffer.create_tag("scored2", foreground_rgba=Gdk.RGBA(0.4, 0, 0, 1))
        self.textbuffer.create_tag("scored3", foreground_rgba=Gdk.RGBA(0.6, 0, 0, 1))
        self.textbuffer.create_tag("scored4", foreground_rgba=Gdk.RGBA(0.8, 0, 0, 1))
        self.textbuffer.create_tag("scored5", foreground_rgba=Gdk.RGBA(1.0, 0, 0, 1))
        self.textbuffer.create_tag("emt", foreground="grey")
        self.textbuffer.create_tag("comment", foreground="#6e71ec")
        self.textbuffer.create_tag(
            "lesson-comment", foreground="green", font_desc=self.font
        )
        self.textbuffer.create_tag("margin", left_margin=4)

        self.selected_tag = self.textbuffer.create_tag(
            "selected",
            background_full_height=True,
            background=self.get_slected_background(),
        )

        # Events
        self.cids_textview = [
            self.textview.connect("motion-notify-event", self.motion_notify_event),
            self.textview.connect("button-press-event", self.button_press_event),
            self.textview.connect("style-updated", self.on_style_updated),
        ]
        self.cid_shown_changed = self.boardview.connect(
            "shownChanged", self.on_shownChanged
        )
        self.cid_remove_variation = self.tag_remove_variation.connect(
            "event", self.tag_event_handler
        )
        self.cids_gamemodel = [
            self.gamemodel.connect_after("game_loaded", self.on_game_loaded),
            self.gamemodel.connect_after("game_changed", self.on_game_changed),
            self.gamemodel.connect_after("game_started", self.update),
            self.gamemodel.connect_after("game_ended", self.update),
            self.gamemodel.connect_after("moves_undone", self.on_moves_undone),
            self.gamemodel.connect_after("variation_undone", self.update),
            self.gamemodel.connect_after("opening_changed", self.update),
            self.gamemodel.connect_after("players_changed", self.on_players_changed),
            self.gamemodel.connect_after("game_terminated", self.on_game_terminated),
            self.gamemodel.connect("variation_added", self.variation_added),
            self.gamemodel.connect("variation_extended", self.variation_extended),
            self.gamemodel.connect("analysis_changed", self.analysis_changed),
            self.gamemodel.connect("analysis_finished", self.update),
        ]

        if self.gamemodel.lesson_game:
            self.cids_gamemodel.append(
                self.gamemodel.connect_after("learn_success", self.on_learn_success)
            )

        # Layout
        __widget__ = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        __widget__.set_spacing(3)
        __widget__.pack_start(self.header_textview, False, False, 0)
        __widget__.pack_start(Gtk.Separator(), False, False, 0)

        self.choices_box = Gtk.Box()
        self.choices_box.connect("realize", add_provider)
        __widget__.pack_start(self.choices_box, False, False, 0)
        self.choices_enabled = True

        sw = Gtk.ScrolledWindow()
        sw.add(self.textview)
        __widget__.pack_start(sw, True, True, 0)

        if self.gamemodel.practice_game or self.gamemodel.lesson_game:
            self.infobar = LearnInfoBar(self.gamemodel, gmwidg.board, self)
            self.boardview.infobar = self.infobar
            __widget__.pack_start(self.infobar, False, False, 0)

        return __widget__

    def fetch_chess_conf(self):
        """
        The method retrieves few parameters from the configuration.
        """
        self.fan = conf.get("figuresInNotation")
        movetext_font = conf.get("movetextFont")
        self.font = Pango.font_description_from_string(movetext_font)
        self.showEmt = conf.get("showEmt")
        self.showBlunder = (
            conf.get("showBlunder") and not self.gamemodel.isPlayingICSGame()
        )
        self.showEval = conf.get("showEval") and not self.gamemodel.isPlayingICSGame()

    def on_game_terminated(self, model):
        """
        The method disconnects all the created events when the widget is destroyed
        at the end of the game.
        """
        for cid in self.cids_textview:
            self.textview.disconnect(cid)
        self.boardview.disconnect(self.cid_shown_changed)
        self.tag_remove_variation.disconnect(self.cid_remove_variation)
        for cid in self.cids_gamemodel:
            self.gamemodel.disconnect(cid)
        for cid in self.cids_conf:
            conf.notify_remove(cid)

    def get_palette(self):
        if isDarkTheme(self.textview):
            palette = [
                "#e5e5e5",
                "#35e119",
                "#ee3e34",
                "#24c6ee",
                "#a882bc",
                "#f09243",
                "#e475e5",
                "#c0c000",
            ]  # white, green, red, aqua, purple, orange, fuchsia, ochre
        else:
            palette = [
                "#4b4b4b",
                "#51a745",
                "#ee3e34",
                "#3965a8",
                "#a882bc",
                "#f09243",
                "#772120",
                "#c0c000",
            ]  # black, green, red, blue, purple, orange, brown, ochre
        return palette

    def get_slected_background(self):
        return "grey" if isDarkTheme(self.textview) else "lightgrey"

    def on_style_updated(self, widget):
        palette = self.get_palette()
        for i in range(64):
            self.tag_vari_depth[i].set_property("foreground", palette[i % len(palette)])

        self.selected_tag.set_property("background", self.get_slected_background())

    def tag_event_handler(self, tag, widget, event, iter):
        """
        The method handles the event specific to a tag, which is further processed
        by the button event of the main widget.
        """
        if (event.type == Gdk.EventType.BUTTON_PRESS) and (
            tag.get_property("name") == "remove-variation"
        ):
            offset = iter.get_offset()
            node = None
            for n in self.nodelist:
                if offset >= n["start"] and offset < n["end"]:
                    node = n
                    break
            if node is None:
                return
            self.variation_to_remove = node
        return False

    def motion_notify_event(self, widget, event):
        """
        The method defines the applicable cursor (standard/hand)
        """
        if self.textview.get_window_type(event.window) not in (
            Gtk.TextWindowType.TEXT,
            Gtk.TextWindowType.PRIVATE,
        ):
            event.window.set_cursor(self.cursor_standard)
            return True

        if event.is_hint:
            # (x, y, state) = event.window.get_pointer()
            (ign, x, y, state) = event.window.get_pointer()
        else:
            x = event.x
            y = event.y
            # state = event.get_state()

        (x, y) = self.textview.window_to_buffer_coords(
            Gtk.TextWindowType.WIDGET, int(x), int(y)
        )

        ret = self.textview.get_iter_at_position(x, y)
        if len(ret) == 3:
            pos_is_over_text, it_at_pos, trailing = ret
        else:
            it_at_pos, trailing = ret

        if it_at_pos.get_child_anchor() is not None:
            event.window.set_cursor(self.cursor_hand)
            return True

        it = self.textview.get_iter_at_location(x, y)

        # https://gramps-project.org/bugs/view.php?id=9335
        if isinstance(it, Gtk.TextIter):
            offset = it.get_offset()
        else:
            offset = it[1].get_offset()

        for node in self.nodelist:
            if offset >= node["start"] and offset < node["end"] and "vari" not in node:
                event.window.set_cursor(self.cursor_hand)
                return True
        event.window.set_cursor(self.cursor_standard)
        return True

    def button_press_event(self, widget, event):
        """
        The method handles the click made on the widget, like showing the board
        editing a comment, or showing a local popup-menu.
        """

        # Detection of the node with the coordinates of the mouse
        (wx, wy) = event.get_coords()
        (x, y) = self.textview.window_to_buffer_coords(
            Gtk.TextWindowType.WIDGET, int(wx), int(wy)
        )
        it = self.textview.get_iter_at_location(x, y)

        # https://gramps-project.org/bugs/view.php?id=9335
        if isinstance(it, Gtk.TextIter):
            offset = it.get_offset()
        else:
            offset = it[1].get_offset()

        node = None
        for n in self.nodelist:
            if offset >= n["start"] and offset < n["end"]:
                node = n
                board = node["board"]
                break
        if node is None:
            return True
        # print("-------------------------------------------------------")
        # print("index is:", self.nodelist.index(node))
        # print(node)
        # print("-------------------------------------------------------")

        # Left click
        if event.button == 1:
            if "vari" in node:
                if self.variation_to_remove is not None:
                    node = self.variation_to_remove
                    self.remove_variation(node)
                    self.gamemodel.remove_variation(node["board"], node["parent"])
                    self.variation_to_remove = None
                return True
            elif "comment" in node:
                self.menu_edit_comment(board=board, index=node["index"])
            else:
                self.boardview.setShownBoard(board.pieceBoard)

        # Right click
        elif event.button == 3:
            self.menu = Gtk.Menu()

            if node is not None:
                position = -1
                for index, child in enumerate(board.children):
                    if isinstance(child, str):
                        position = index
                        break

                menuitem = Gtk.MenuItem(_("Refresh"))
                menuitem.connect("activate", self.menu_refresh)
                self.menu.append(menuitem)

                if (
                    len(self.gamemodel.boards) > 1
                    and board == self.gamemodel.boards[1].board
                    and not self.gamemodel.boards[0].board.children
                ):
                    menuitem = Gtk.MenuItem(_("Add start comment"))
                    menuitem.connect(
                        "activate",
                        self.menu_edit_comment,
                        self.gamemodel.boards[0].board,
                        0,
                    )
                    self.menu.append(menuitem)

                if position == -1:
                    menuitem = Gtk.MenuItem(_("Add comment"))
                    menuitem.connect("activate", self.menu_edit_comment, board, 0)
                else:
                    menuitem = Gtk.MenuItem(_("Edit comment"))
                    menuitem.connect(
                        "activate", self.menu_edit_comment, board, position
                    )
                self.menu.append(menuitem)

                symbol_menu1 = Gtk.Menu()
                for nag, menutext in (
                    ("$1", _("Good move")),
                    ("$2", _("Bad move")),
                    ("$3", _("Excellent move")),
                    ("$4", _("Very bad move")),
                    ("$5", _("Interesting move")),
                    ("$6", _("Suspicious move")),
                    ("$7", _("Forced move")),
                ):
                    menuitem = Gtk.MenuItem(f"{nag2symbol(nag)} {menutext}")
                    menuitem.connect("activate", self.menu_move_attribute, board, nag)
                    symbol_menu1.append(menuitem)

                menuitem = Gtk.MenuItem(_("Add move symbol"))
                menuitem.set_submenu(symbol_menu1)
                self.menu.append(menuitem)

                symbol_menu2 = Gtk.Menu()
                for nag, menutext in (
                    ("$10", _("Drawish")),
                    ("$13", _("Unclear position")),
                    ("$14", _("Slight advantage")),
                    ("$16", _("Moderate advantage")),
                    ("$18", _("Decisive advantage")),
                    ("$20", _("Crushing advantage")),
                    ("$22", _("Zugzwang")),
                    ("$32", _("Development advantage")),
                    ("$36", _("Initiative")),
                    ("$40", _("With attack")),
                    ("$44", _("Compensation")),
                    ("$132", _("Counterplay")),
                    ("$138", _("Time pressure")),
                ):
                    menuitem = Gtk.MenuItem(f"{nag2symbol(nag)} {menutext}")
                    menuitem.connect(
                        "activate", self.menu_position_attribute, board, nag
                    )
                    symbol_menu2.append(menuitem)

                menuitem = Gtk.MenuItem(_("Add evaluation symbol"))
                menuitem.set_submenu(symbol_menu2)
                self.menu.append(menuitem)

                self.menu.append(Gtk.SeparatorMenuItem())

                removals_menu = Gtk.Menu()

                menuitem = Gtk.MenuItem(_("Comment"))
                menuitem.connect("activate", self.menu_delete_comment, board, position)
                removals_menu.append(menuitem)

                menuitem = Gtk.MenuItem(_("Symbols"))
                menuitem.connect("activate", self.menu_remove_symbols, board)
                removals_menu.append(menuitem)

                menuitem = Gtk.MenuItem(_("All the evaluations"))
                menuitem.connect("activate", self.menu_reset_evaluations)
                removals_menu.append(menuitem)

                menuitem = Gtk.MenuItem(_("Remove"))
                menuitem.set_submenu(removals_menu)
                self.menu.append(menuitem)

                self.menu.show_all()
                self.menu.popup(None, None, None, None, event.button, event.time)
        return True

    def menu_refresh(self, widget):
        self.update()

    def menu_edit_comment(self, widget=None, board=None, index=0):
        """
        The method will create/update or delete a comment.
        The popup window will receive an additional button if there is an existing comment.
        """
        creation = True
        if not board.children:
            board.children.append("")
        elif not isinstance(board.children[index], str):
            board.children.insert(index, "")
        else:
            creation = False

        buttons_list = () if creation else (Gtk.STOCK_CLEAR, Gtk.ResponseType.REJECT)
        buttons_list = buttons_list + (
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL,
            Gtk.STOCK_OK,
            Gtk.ResponseType.ACCEPT,
        )

        dialog = Gtk.Dialog(
            _("Add comment") if creation else _("Edit comment"),
            mainwindow(),
            Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
            buttons_list,
        )

        textedit = Gtk.TextView()
        textedit.set_editable(True)
        textedit.set_cursor_visible(True)
        textedit.set_wrap_mode(Gtk.WrapMode.WORD)
        textedit.set_accepts_tab(False)

        textbuffer = textedit.get_buffer()
        textbuffer.set_text(board.children[index])

        sw = Gtk.ScrolledWindow()
        sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        sw.add(textedit)

        dialog.get_content_area().pack_start(sw, True, True, 0)
        dialog.resize(300, 200)
        dialog.show_all()

        response = dialog.run()
        dialog.destroy()
        if response == Gtk.ResponseType.DELETE_EVENT:  # Escape key implies Cancel
            response = Gtk.ResponseType.CANCEL

        (iter_first, iter_last) = textbuffer.get_bounds()
        comment = (
            ""
            if response == Gtk.ResponseType.REJECT
            else textbuffer.get_text(iter_first, iter_last, False)
        )
        if response != Gtk.ResponseType.CANCEL and board.children[index] != comment:
            if len(comment) == 0:
                del board.children[index]
            else:
                board.children[index] = comment
            self.gamemodel.needsSave = True
            self.update()

    def menu_delete_comment(self, widget=None, board=None, index=0):
        """
        The method removes a comment.
        """
        if index == -1 or not board.children:
            return
        elif not isinstance(board.children[index], str):
            return
        self.gamemodel.needsSave = True
        del board.children[index]
        self.update()

    def menu_move_attribute(self, widget, board, nag):
        """
        The method will assign a sign to a move, like "a4!" or "Kh8?!".
        It is not possible to have multiple NAG tags for the move.
        """
        if len(board.nags) == 0:
            board.nags.append(nag)
            self.gamemodel.needsSave = True
        else:
            if board.nags[0] != nag:
                board.nags[0] = nag
                self.gamemodel.needsSave = True

        if self.gamemodel.needsSave:
            self.update_node(board)

    def menu_position_attribute(self, widget, board, nag):
        """
        The method will assign a sign to describe the current position.
        It is not possible to have multiple NAG tags for the position.
        """
        color = board.color
        if color == WHITE and nag in (
            "$14",
            "$16",
            "$18",
            "$20",
            "$22",
            "$32",
            "$36",
            "$40",
            "$44",
            "$132",
            "$138",
        ):
            nag = "$%s" % (int(nag[1:]) + 1)

        if len(board.nags) == 0:
            board.nags.append("")
            board.nags.append(nag)
            self.gamemodel.needsSave = True
        if len(board.nags) == 1:
            board.nags.append(nag)
            self.gamemodel.needsSave = True
        else:
            if board.nags[1] != nag:
                board.nags[1] = nag
                self.gamemodel.needsSave = True

        if self.gamemodel.needsSave:
            self.update_node(board)

    def menu_remove_symbols(self, widget, board):
        """
        The method removes all the NAG tags assigned to the current node.
        """
        if board.nags:
            board.nags = []
            self.update_node(board)
            self.gamemodel.needsSave = True

    def menu_reset_evaluations(self, widget):
        """
        The method removes all the evaluations in order to recalculate them
        or simplify the output.
        """
        self.gamemodel.scores = {}
        self.update()

    def print_node(self, node):
        """Just a debug helper"""

        if "vari" in node:
            if node["board"] is None:
                text = "["
            else:
                text = "]"
        elif "comment" in node:
            text = node["comment"]
        else:
            text = self.__movestr(node["board"])
        return text

    def remove_variation(self, node):
        parent = node["parent"]

        # Set new shown board to parent board by default
        if parent.pieceBoard is None:
            # variation without played move at game end
            self.boardview.setShownBoard(self.gamemodel.boards[-1])
        else:
            self.boardview.setShownBoard(parent.pieceBoard)

        last_node = self.nodelist[-1] == node

        startnode = node["vari"]
        start = startnode["start"]
        end = node["end"]

        need_delete = []
        for n in self.nodelist[self.nodelist.index(startnode) :]:
            if n["start"] < end:
                need_delete.append(n)

        if not last_node:
            diff = end - start
            for n in self.nodelist[self.nodelist.index(node) + 1 :]:
                n["start"] -= diff
                n["end"] -= diff

        for n in need_delete:
            self.nodelist.remove(n)

        start_iter = self.textbuffer.get_iter_at_offset(start)
        end_iter = self.textbuffer.get_iter_at_offset(end)
        self.textbuffer.delete(start_iter, end_iter)

    def variation_start(self, iter, index, level):
        start = iter.get_offset()
        if not iter.ends_tag(tag=self.tag_new_line):
            self.textbuffer.insert_with_tags_by_name(iter, "\n", "new_line")
        vlevel = min(level + 1, len(self.tag_vari_depth) - 1)
        self.textbuffer.insert_with_tags_by_name(
            iter, "(", "variation-depth-%d" % vlevel
        )

        node = {}
        node["board"] = EMPTY_BOARD
        node["vari"] = None
        node["start"] = start
        node["end"] = iter.get_offset()

        if index == -1:
            self.nodelist.append(node)
        else:
            self.nodelist.insert(index, node)

        return (iter.get_offset() - start, node)

    def variation_end(self, iter, index, level, firstboard, parent, opening_node):
        start = iter.get_offset()
        vlevel = min(level + 1, len(self.tag_vari_depth) - 1)
        self.textbuffer.insert_with_tags_by_name(
            iter, ")", "variation-depth-%d" % vlevel
        )

        self.textbuffer.insert_with_tags_by_name(iter, " ✖ ", "remove-variation")
        # chr = iter.get_char()

        # somehow iter.begins_tag() doesn't work, so we use get_char() instead
        if iter.get_char() != "\n":
            self.textbuffer.insert_with_tags_by_name(iter, "\n", "new_line")

        node = {}
        node["board"] = firstboard
        node["parent"] = parent
        node["vari"] = opening_node
        node["start"] = start
        node["end"] = iter.get_offset()

        if index == -1:
            self.nodelist.append(node)
        else:
            self.nodelist.insert(index, node)

        return iter.get_offset() - start

    def update_node(self, board):
        """Called after adding/removing evaluation symbols"""
        node = None
        for n in self.nodelist:
            if n["board"] == board:
                start = self.textbuffer.get_iter_at_offset(n["start"])
                end = self.textbuffer.get_iter_at_offset(n["end"])
                node = n
                break
        if node is None:
            return

        index = self.nodelist.index(node)
        level = node["level"]
        parent = node["parent"]
        diff = node["end"] - node["start"]
        self.nodelist.remove(node)
        self.textbuffer.delete(start, end)
        inserted_node = self.insert_node(board, start, index, level, parent)
        diff = inserted_node["end"] - inserted_node["start"] - diff

        if len(self.nodelist) > index + 1:
            for node in self.nodelist[index + 1 :]:
                node["start"] += diff
                node["end"] += diff
        self.update_selected_node()

    def insert_node(self, board, iter, index, level, parent):
        start = iter.get_offset()
        movestr = self.__movestr(board)
        self.textbuffer.insert(iter, "%s " % movestr)

        startIter = self.textbuffer.get_iter_at_offset(start)
        endIter = self.textbuffer.get_iter_at_offset(iter.get_offset())

        if level == 0:
            self.textbuffer.apply_tag_by_name("move", startIter, endIter)
            self.textbuffer.apply_tag_by_name("margin", startIter, endIter)
            self.colorize_node(board.plyCount, startIter, endIter)
        else:
            self.textbuffer.apply_tag_by_name(
                "variation-depth-%d" % level, startIter, endIter
            )

        node = {}
        node["board"] = board
        node["start"] = start
        node["end"] = iter.get_offset()
        node["parent"] = parent
        node["level"] = level

        if index == -1:
            self.nodelist.append(node)
        else:
            self.nodelist.insert(index, node)
        return node

    def variation_extended(self, gamemodel, prev_board, board):
        node = None
        for n in self.nodelist:
            if n["board"] == prev_board:
                end = self.textbuffer.get_iter_at_offset(n["end"])
                node = n
                break

        node_index = self.nodelist.index(node) + 1

        inserted_node = self.insert_node(
            board, end, node_index, node["level"], node["parent"]
        )
        diff = inserted_node["end"] - inserted_node["start"]

        if len(self.nodelist) > node_index + 1:
            for node in self.nodelist[node_index + 1 :]:
                node["start"] += diff
                node["end"] += diff

        self.boardview.setShownBoard(board.pieceBoard)
        self.gamemodel.needsSave = True

    def hide_movelist(self):
        return (self.gamemodel.lesson_game and not self.gamemodel.solved) or (
            self.gamemodel.puzzle_game and len(self.gamemodel.moves) == 0
        )

    def variation_added(self, gamemodel, boards, parent):
        # Don't show moves in interactive lesson games
        if self.hide_movelist():
            return

        # first find the iter where we will inset this new variation
        node = None
        for n in self.nodelist:
            if n["board"] == parent:
                end = self.textbuffer.get_iter_at_offset(n["end"])
                node = n
                break

        if node is None:
            next_node_index = len(self.nodelist)
            end = self.textbuffer.get_end_iter()
            level = 0
        else:
            next_node_index = self.nodelist.index(node) + 1
            level = node["level"]

        # diff will store the offset we need to shift the remaining stuff
        diff = 0

        # variation opening parenthesis
        sdiff, opening_node = self.variation_start(end, next_node_index, level)
        diff += sdiff

        for i, board in enumerate(boards):
            # do we have initial variation comment?
            if board.prev is None:
                continue
            else:
                # insert variation move
                inserted_node = self.insert_node(
                    board, end, next_node_index + i, level + 1, parent
                )
                diff += inserted_node["end"] - inserted_node["start"]
                end = self.textbuffer.get_iter_at_offset(inserted_node["end"])

        diff += self.variation_end(
            end, next_node_index + len(boards), level, boards[1], parent, opening_node
        )

        # adjust remaining stuff offsets
        if next_node_index > 0:
            for node in self.nodelist[next_node_index + len(boards) + 1 :]:
                node["start"] += diff
                node["end"] += diff

        # if new variation is coming from clicking in book panel
        # we want to jump into the first board in new vari
        self.boardview.setShownBoard(boards[1].pieceBoard)
        self.gamemodel.needsSave = True

    def colorize_node(self, ply, start, end):
        """
        The method updates the color or the node in order to show the errors and blunders.
        """
        tags = ["emt", "scored0", "scored5", "scored4", "scored3", "scored2", "scored1"]
        tags_diff = [None, None, 400, 200, 90, 50, 20]
        for tag_name in tags:
            self.textbuffer.remove_tag_by_name(tag_name, start, end)

        tag_name = "scored0"
        if (
            self.showBlunder
            and ply - 1 in self.gamemodel.scores
            and ply in self.gamemodel.scores
        ):
            color = (ply - 1) % 2
            oldmoves, oldscore, olddepth = self.gamemodel.scores[ply - 1]
            oldscore = oldscore * -1 if color == BLACK else oldscore
            moves, score, depth = self.gamemodel.scores[ply]
            score = score * -1 if color == WHITE else score
            diff = score - oldscore
            for i, td in enumerate(tags_diff):
                if td is None:
                    continue
                if (diff >= td and color == BLACK) or (diff <= -td and color == WHITE):
                    tag_name = tags[i]
                    break
        self.textbuffer.apply_tag_by_name(tag_name, start, end)

    def analysis_changed(self, gamemodel, ply):
        """
        The method updates the analysis received from an external event.
        """
        if self.boardview.animating:
            return
        if not self.boardview.shownIsMainLine():
            return

        try:
            board = gamemodel.getBoardAtPly(ply).board
        except IndexError:
            return

        node = None
        if self.showEval or self.showBlunder:
            for n in self.nodelist:
                if n["board"] == board:
                    start = self.textbuffer.get_iter_at_offset(n["start"])
                    end = self.textbuffer.get_iter_at_offset(n["end"])
                    node = n
                    break
        if node is None:
            return

        if self.showBlunder:
            self.colorize_node(ply, start, end)

        emt_eval = ""
        if self.showEmt and self.gamemodel.timemodel.hasTimes:
            elapsed = gamemodel.timemodel.getElapsedMoveTime(
                board.plyCount - gamemodel.lowply
            )
            emt_eval = "%s " % formatTime(elapsed)

        if self.showEval:
            if board.plyCount in gamemodel.scores:
                moves, score, depth = gamemodel.scores[board.plyCount]
                score = score * -1 if board.color == BLACK else score
                emt_eval += "%s " % prettyPrintScore(score, depth, format_mate=True)

        if emt_eval:
            if node == self.nodelist[-1]:
                next_node = None
                self.textbuffer.delete(end, self.textbuffer.get_end_iter())
            else:
                next_node = self.nodelist[self.nodelist.index(node) + 1]
                next_start = self.textbuffer.get_iter_at_offset(next_node["start"])
                self.textbuffer.delete(end, next_start)
            self.textbuffer.insert_with_tags_by_name(end, emt_eval, "emt")

            if next_node is not None:
                diff = end.get_offset() - next_node["start"]
                for node in self.nodelist[self.nodelist.index(next_node) :]:
                    node["start"] += diff
                    node["end"] += diff

    def update_selected_node(self):
        """Update the selected node highlight"""
        self.textbuffer.remove_tag_by_name(
            "selected", self.textbuffer.get_start_iter(), self.textbuffer.get_end_iter()
        )
        shown_board = self.gamemodel.getBoardAtPly(
            self.boardview.shown, self.boardview.shown_variation_idx
        )
        start = None
        for node in self.nodelist:
            if node["board"] == shown_board.board:
                start = self.textbuffer.get_iter_at_offset(node["start"])
                end = self.textbuffer.get_iter_at_offset(node["end"])
                # don't hightlight initial game comment!
                if shown_board.board != self.gamemodel.boards[0].board:
                    self.textbuffer.apply_tag_by_name("selected", start, end)
                break

        if start:
            # self.textview.scroll_to_iter(start, within_margin=0.03)
            self.textview.scroll_to_iter(start, 0.01, True, 0.00, 0.00)

    def insert_nodes(self, board, level=0, parent=None, result=None):
        """Recursively builds the node tree"""

        # Don't show moves in interactive lesson games
        if self.hide_movelist():
            return

        end_iter = self.textbuffer.get_end_iter  # Convenience shortcut to the function

        while True:
            # start = end_iter().get_offset()

            if board is None:
                break

            # Initial game or variation comment
            if board.prev is None:
                for index, child in enumerate(board.children):
                    if isinstance(child, str):
                        self.insert_comment(
                            child,
                            board,
                            parent,
                            index=index,
                            level=level,
                            ini_board=board,
                        )
                board = board.next
                continue

            if board.fen_was_applied:
                self.insert_node(board, end_iter(), -1, level, parent)

            if (
                self.showEmt
                and level == 0
                and board.fen_was_applied
                and self.gamemodel.timemodel.hasTimes
            ):
                elapsed = self.gamemodel.timemodel.getElapsedMoveTime(
                    board.plyCount - self.gamemodel.lowply
                )
                self.textbuffer.insert_with_tags_by_name(
                    end_iter(), "%s " % formatTime(elapsed), "emt"
                )

            if (
                self.showEval
                and level == 0
                and board.fen_was_applied
                and board.plyCount in self.gamemodel.scores
            ):
                moves, score, depth = self.gamemodel.scores[board.plyCount]
                score = score * -1 if board.color == BLACK else score
                # endIter = self.textbuffer.get_iter_at_offset(end_iter().get_offset())
                self.textbuffer.insert_with_tags_by_name(
                    end_iter(),
                    "%s " % prettyPrintScore(score, depth, format_mate=True),
                    "emt",
                )

            for index, child in enumerate(board.children):
                if isinstance(child, str):
                    # comment
                    self.insert_comment(child, board, parent, index=index, level=level)
                else:
                    # variation
                    diff, opening_node = self.variation_start(end_iter(), -1, level)
                    self.insert_nodes(child[0], level + 1, parent=board)
                    self.variation_end(
                        end_iter(), -1, level, child[1], board, opening_node
                    )

            if board.next:
                board = board.next
            else:
                break

        if (
            result
            and result != "*"
            and not self.gamemodel.lesson_game
            and not self.gamemodel.practice_game
        ):
            self.textbuffer.insert_with_tags_by_name(end_iter(), " " + result, "move")

    def apply_symbols(self, text):
        """
        The method will apply a Unicode symbol for any move contained in a sentence.
        Because it applies to a PGN-compatible text, only English letters are replaced (RNBQK).
        """

        def process_word(word):
            # Undecoration of the word
            regex = re_decoration.search(word)
            if regex:
                lead, core, trail = regex.groups()

                # Detection of the pieces in the move
                regex = re_move.search(core)
                if regex:
                    parts = list(regex.groups())

                    # Application of the Unicode symbols
                    for i, sign in enumerate(
                        reprSign
                    ):  # TODO what about reprSignMakruk and reprSignSittuyin ?
                        if parts[0] == sign:
                            parts[0] = FAN_PIECES[WHITE][i]
                        if parts[2] == sign:
                            parts[2] = FAN_PIECES[WHITE][i]

                    return lead + "".join(parts) + trail
                else:
                    return word
            else:
                return word

        # Application of the filter on each element of the text
        if self.fan:
            re_decoration = re.compile(
                r"^([^a-hprnkqx1-8]*|[0-9]+\.+)?([a-hprnkqx1-8=@]+)([^a-hprnkqx1-8]*)$",
                re.IGNORECASE,
            )
            re_move = re.compile(
                "^([PRNBQK]?)(@?[a-h]?[1-8]?x?[a-h][1-8]=?)([RNBQK]?)(.*)$"
            )
            return " ".join([process_word(word) for word in text.split(" ")])
        else:
            return text

    def insert_comment(self, comment, board, parent, index=0, level=0, ini_board=None):
        comment = re.sub(r"\[%.*?\]", "", comment)
        if not comment:
            return

        end_iter = self.textbuffer.get_end_iter()
        pos = len(self.nodelist)
        for n in self.nodelist:
            if n["board"] == board:
                pos = self.nodelist.index(n)
                break
        start = end_iter.get_offset()

        self.textbuffer.insert_with_tags_by_name(
            end_iter, self.apply_symbols(comment) + " ", "comment"
        )

        node = {}
        node["board"] = ini_board if ini_board is not None else board
        node["comment"] = comment
        node["index"] = index
        node["parent"] = parent
        node["level"] = level
        node["start"] = start
        node["end"] = end_iter.get_offset()
        self.nodelist.insert(pos if ini_board is not None else pos + 1, node)

        return node

    def update_header(self):
        self.header_textbuffer.set_text("")

        end_iter = self.header_textbuffer.get_end_iter

        if self.gamemodel.info is not None:
            insert_formatted(self.header_textview, end_iter(), self.gamemodel.info)
            self.header_textbuffer.insert(end_iter(), "\n")

        if self.gamemodel.players:
            text = repr(self.gamemodel.players[0])
        else:
            return

        self.header_textbuffer.insert_with_tags_by_name(end_iter(), text, "head2")
        white_elo = self.gamemodel.tags["WhiteElo"]
        if white_elo not in [None, "", "?", "0", 0]:
            self.header_textbuffer.insert_with_tags_by_name(
                end_iter(), " %s" % white_elo, "head1"
            )

        self.header_textbuffer.insert_with_tags_by_name(end_iter(), " - ", "head1")

        # text = self.gamemodel.tags['Black']
        text = repr(self.gamemodel.players[1])
        self.header_textbuffer.insert_with_tags_by_name(end_iter(), text, "head2")
        black_elo = self.gamemodel.tags["BlackElo"]
        if black_elo not in [None, "", "?", "0", 0]:
            self.header_textbuffer.insert_with_tags_by_name(
                end_iter(), " %s" % black_elo, "head1"
            )

        result = reprResult[self.gamemodel.status]
        self.header_textbuffer.insert_with_tags_by_name(
            end_iter(), " " + result + "\n", "head2"
        )

        text = ""
        time_control = self.gamemodel.tags.get("TimeControl")
        if time_control:
            match = parseTimeControlTag(time_control)
            if match is None:
                text += _("No time control") if time_control == "-" else time_control
            else:
                secs, inc, moves = match

                ttime = ""
                tmin = int(floor(secs / 60))
                tsec = secs - 60 * tmin
                if tmin > 0:
                    ttime += str(tmin) + " " + (_("mins") if tmin > 1 else _("min"))
                if tsec > 0:
                    if ttime != "":
                        ttime += " "
                    ttime += str(tsec) + " " + (_("secs") if tsec > 1 else _("sec"))

                if moves is not None and moves > 0:
                    text += _("%(time)s for %(count)d moves") % (
                        {"time": ttime, "count": moves}
                    )
                else:
                    text += ttime
                    if inc != 0:
                        text += (
                            (" + " if inc >= 0 else " – ")
                            + str(abs(inc))
                            + " "
                            + (_("secs") if abs(inc) > 1 else _("sec"))
                            + "/"
                            + _("move")
                        )

        event = self.gamemodel.tags["Event"]
        if event and event != "?":
            if len(text) > 0:
                text += ", "
            text += event

        site = self.gamemodel.tags["Site"]
        if site and site != "?":
            if len(text) > 0:
                text += ", "
            text += site

        round = self.gamemodel.tags["Round"]
        if round and round != "?":
            if len(text) > 0:
                text += ", "
            text += _("round %s") % round

        date = self.gamemodel.tags["Date"]
        date = date.replace(".??", "").replace("????.", "")
        if date != "":
            if len(text) > 0:
                text += ", "
            text += date
        self.header_textbuffer.insert_with_tags_by_name(end_iter(), text, "head1")

        eco = self.gamemodel.tags.get("ECO")
        if eco:
            self.header_textbuffer.insert_with_tags_by_name(
                end_iter(), "\n" + eco, "head2"
            )
            opening = self.gamemodel.tags.get("Opening")
            if opening:
                self.header_textbuffer.insert_with_tags_by_name(
                    end_iter(), " - ", "head1"
                )
                self.header_textbuffer.insert_with_tags_by_name(
                    end_iter(), opening, "head2"
                )
            variation = self.gamemodel.tags.get("Variation")
            if variation:
                self.header_textbuffer.insert_with_tags_by_name(
                    end_iter(), ", ", "head1"
                )
                self.header_textbuffer.insert_with_tags_by_name(
                    end_iter(), variation, "head2"
                )

    def update(self, *args):
        """
        This method execute the full refresh of the widget.
        """
        self.fetch_chess_conf()
        self.textbuffer.set_text("")
        self.nodelist = []
        self.update_header()
        self.update_choices()

        result = reprResult[self.gamemodel.status]

        self.insert_nodes(self.gamemodel.boards[0].board, result=result)
        self.update_selected_node()

    def on_choice_clicked(self, button, board):
        self.boardview.setShownBoard(board)
        if self.gamemodel.lesson_game:
            self.infobar.opp_choice_selected(board)

    def on_enter_notify_event(self, button, event, move):
        arrow = Cord(FCORD(move), color="G"), Cord(TCORD(move))
        self.boardview.arrows.add(arrow)
        self.boardview.redrawCanvas()

    def on_leave_notify_event(self, button, event, move):
        arrow = Cord(FCORD(move), color="G"), Cord(TCORD(move))
        if arrow in self.boardview.arrows:
            self.boardview.arrows.remove(arrow)
            self.boardview.redrawCanvas()

    def remove_choices(self):
        """Removes all choice buttons"""
        for widget in self.choices_box:
            self.choices_box.remove(widget)

    def update_choices(self):
        # First update lesson move comments
        if self.hide_movelist():
            self.show_lesson_comments()

        view = self.boardview
        try:
            next_board = view.model.getBoardAtPly(
                view.shown + 1, variation=view.shown_variation_idx
            )
        except IndexError:
            next_board = None

        # On game end and variation end there will be no choices for sure
        if next_board is None:
            self.remove_choices()
            return

        base_board = view.model.getBoardAtPly(
            view.shown, variation=view.shown_variation_idx
        )

        # Don't show our choices in lesson games
        if (
            self.gamemodel.lesson_game
            and base_board.color == self.gamemodel.orientation
        ):
            self.remove_choices()
            return

        # Gether variations first moves if there are any
        choices = []
        for child in next_board.board.children:
            if isinstance(child, list):
                lboard = child[1]
                if self.fan:
                    text = toFAN(base_board.board, lboard.lastMove)
                else:
                    text = toSAN(base_board.board, lboard.lastMove, True)
                choices.append((lboard.pieceBoard, lboard.lastMove, text))

        # Add main line next move to choice list also
        if choices:
            move = next_board.board.lastMove
            if self.fan:
                text = toFAN(base_board.board, move)
            else:
                text = toSAN(base_board.board, move, True)
            choices = [(next_board, move, text)] + choices

        # Remove previous choice buttons
        self.remove_choices()

        # Add nev choice buttons
        if self.choices_enabled and choices:
            for board, move, san in choices:
                button = Gtk.Button(san)
                button.set_name("rounded")
                button.connect("clicked", self.on_choice_clicked, board)
                button.connect("enter_notify_event", self.on_enter_notify_event, move)
                button.connect("leave_notify_event", self.on_leave_notify_event, move)
                self.choices_box.pack_start(button, False, False, 3)
            self.choices_box.show_all()
            preferencesDialog.SoundTab.playAction("variationChoice")

        if not self.choices_enabled:
            self.choices_enabled = True

    def show_lesson_comments(self):
        self.textbuffer.set_text("")
        self.nodelist = []
        view = self.boardview
        board = view.model.getBoardAtPly(view.shown, variation=view.shown_variation_idx)
        for index, child in enumerate(board.board.children):
            if isinstance(child, str):
                if child.lstrip().startswith("[%"):
                    continue
                end_iter = self.textbuffer.get_end_iter()
                self.textbuffer.insert_with_tags_by_name(
                    end_iter, self.apply_symbols(child) + " ", "lesson-comment"
                )

    def on_shownChanged(self, view, shown):
        self.update_choices()
        self.update_selected_node()

    def on_moves_undone(self, game, moves):
        """
        This method is called once a move has been undone.
        """
        start = self.textbuffer.get_start_iter()
        end = self.textbuffer.get_end_iter()
        for node in reversed(self.nodelist):
            if node["board"].pieceBoard == self.gamemodel.variations[0][-1]:
                start = self.textbuffer.get_iter_at_offset(node["end"])
                break
            else:
                self.nodelist.remove(node)

        if self.gamemodel.ply > 0:
            self.textbuffer.delete(start, end)
        self.update()

    def on_learn_success(self, model):
        self.update()

    def on_players_changed(self, model):
        self.update_header()

    def on_game_loaded(self, model, uri):
        """
        The method is called when a game is loaded.
        """
        if model.tags.get("ECO", "") == "" and model.tags.get("Opening", "") == "":
            model.setOpening(redetermine=True)
        self.update()

    def on_game_changed(self, game, ply):
        """
        The method is called when a game is changed, like by a move.
        """
        if self.hide_movelist():
            self.update()

        board = game.getBoardAtPly(ply, variation=0).board
        # if self.update() inserted all nodes before (f.e opening_changed), do nothing
        if self.nodelist and self.nodelist[-1]["board"] == board:
            return
        end_iter = self.textbuffer.get_end_iter
        start = end_iter().get_offset()
        movestr = self.__movestr(board)
        self.textbuffer.insert(end_iter(), "%s " % movestr)

        startIter = self.textbuffer.get_iter_at_offset(start)
        endIter = self.textbuffer.get_iter_at_offset(end_iter().get_offset())

        self.textbuffer.apply_tag_by_name("move", startIter, endIter)
        self.colorize_node(board.plyCount, startIter, endIter)

        node = {}
        node["board"] = board
        node["start"] = startIter.get_offset()
        node["end"] = end_iter().get_offset()
        node["parent"] = None
        node["level"] = 0

        self.nodelist.append(node)

        if self.showEmt and self.gamemodel.timed:
            elapsed = self.gamemodel.timemodel.getElapsedMoveTime(
                board.plyCount - self.gamemodel.lowply
            )
            self.textbuffer.insert_with_tags_by_name(
                end_iter(), "%s " % formatTime(elapsed), "emt"
            )

        self.update_selected_node()

    def __movestr(self, board):
        move = board.lastMove
        if self.fan:
            movestr = toFAN(board.prev, move)
        else:
            movestr = toSAN(board.prev, move, True)
        nagsymbols = "".join([nag2symbol(nag) for nag in board.nags])
        # To prevent wrap castling we will use hyphen bullet (U+2043)
        return "{}{}{}".format(move_count(board), movestr.replace("-", "⁃"), nagsymbols)



# Atomic Chess

from pychess.Utils.const import (
    VARIANTS_OTHER_NONSTANDARD,
    KING,
    ATOMICCHESS,
    ENPASSANT,
    B8,
    E1,
)
from pychess.Utils.Board import Board
from pychess.Utils.Cord import Cord
from pychess.Utils.Move import Move
from pychess.Utils.lutils.bitboard import iterBits
from pychess.Utils.lutils.ldata import moveArray


class AtomicBoard(Board):
    variant = ATOMICCHESS
    __desc__ = _("FICS atomic: http://www.freechess.org/Help/HelpFiles/atomic.html")
    name = _("Atomic")
    cecp_name = "atomic"
    need_initial_board = False
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD


def cordsAround(cord):
    kingMoves = moveArray[KING]
    for co_ord in iterBits(kingMoves[cord.cord]):
        yield Cord(co_ord)


def piecesAround(board, cord):
    kingMoves = moveArray[KING]

    friends = board.friends[board.color]
    for co_ord in iterBits(kingMoves[cord] & friends):
        yield co_ord, board.arBoard[co_ord], board.color

    enemies = board.friends[1 - board.color]
    for co_ord in iterBits(kingMoves[cord] & enemies):
        yield co_ord, board.arBoard[co_ord], 1 - board.color


def kingExplode(board, move, color):
    tcord = move & 63
    # fcord = (move >> 6) & 63
    flag = move >> 12
    if board.arBoard[tcord] or flag == ENPASSANT:
        for acord, apiece, acolor in piecesAround(board, tcord):
            if apiece == KING and acolor == color:
                return True
    return False


if __name__ == "__main__":
    FEN = "rnbqkbnr/ppp1pppp/8/8/8/8/PPPPPPPP/RNBQKBNR b KQkq - 0 1"
    atomic_board = AtomicBoard(FEN)
    print(atomic_board.board.__repr__())

    for acord, apiece, acolor in piecesAround(atomic_board.board, B8):
        print(acord, apiece, acolor)
    for acord, apiece, acolor in piecesAround(atomic_board.board, E1):
        print(acord, apiece, acolor)

    from pychess.Utils.lutils.lmove import parseAN

    atomic_board = atomic_board.move(Move(parseAN(atomic_board.board, "d8d2")))
    print(atomic_board.board.__repr__())
    print(atomic_board.board.pieceCount)

    atomic_board.board.popMove()
    print(atomic_board.board.__repr__())
    print(atomic_board.board.pieceCount)
